"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

__all__ = ['husimi_2d','husimi_3d','wigner_2d','wigner_3d',
        'husimi_spin_3d','wigner_spin_3d','cmindex']

from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib import cm
from numpy import amax, meshgrid
import numpy as np
from tqix.quasi_prob import *
import os 

def husimi_2d(state,xrange,yrange,N = 100,fname='fig_husimi_2d.eps',
            cmap = 'viridis',alpha = 1.0):
    """
    to visualize a Husimi Q function

    Parameters:
    ------------
    state: quantum object
        A given quantum state needed to visualize
    xrange, yrange: array-like(2)
        The minimum and maximum values of the coordinates 
    N: integer
        number of steps for xrange, yrange
    fname: string
        File name  
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 

    Returns:
    A file with fname
    """

    xarray = np.linspace(xrange[0],xrange[1],N)
    yarray = np.linspace(yrange[0],yrange[1],N)
    zarray = husimi(state,xarray,yarray)
    zarray /= amax(zarray)

    fig,axes = plt.subplots(1,1,figsize=(6,6))
    cont = axes.contourf(xarray,yarray,zarray,80,cmap = cmap,alpha = alpha)
    plt.xlabel("x")
    plt.ylabel("y")

    for c in cont.collections:
        c.set_edgecolor("face")

    _printout(fname)
    plt.savefig(fname, dpi=25)

def husimi_3d(state,xrange,yrange,N = 100,fname='fig_husimi_3d.eps',
            cmap = 'viridis',alpha = 1.0):
    """
    to visualize a 3d Husimi function

    Parameters:
    ------------
    state: quantum object
        A given quantum state needed to visualize
    xrange, yrange: array-like(2)
        The minimum and maximum values of the coordinates 
    N: integer
        number of steps for xrange, yrange
    fname: string
        File name  
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 

    Returns:
    A file with fname
    """
    xarray = np.linspace(xrange[0],xrange[1],N)
    yarray = np.linspace(yrange[0],yrange[1],N)
    zarray = husimi(state,xarray,yarray)
    zarray /= amax(zarray)

    xx, yy = meshgrid(xarray,yarray)

    norm = plt.Normalize(zarray.min(), zarray.max())
    colors = cm.viridis(norm(zarray))

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.plot_surface(xx, yy, zarray, cmap=cmap,
                    rstride=1, cstride=1,linewidth=0,facecolors=colors)

    plt.xlabel("x")
    plt.ylabel("y")

    _printout(fname)
    plt.savefig(fname, dpi=25)

def wigner_2d(state,xrange,yrange,N = 100,fname='fig_wigner_2d.eps',
            cmap = 'viridis',alpha = 1.0):

    """
    to visualize a Wigner function
        Parameters:
    ------------
    state: quantum object
        A given quantum state needed to visualize
    xrange, yrange: array-like(2)
        The minimum and maximum values of the coordinates 
    N: integer
        number of steps for xrange, yrange
    fname: string
        File name  
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 

    Returns:
    A file with fname
    """

    xarray = np.linspace(xrange[0],xrange[1],N)
    yarray = np.linspace(yrange[0],yrange[1],N)
    zarray = wigner(state,xarray,yarray)
    zarray /= amax(zarray)

    fig,axes = plt.subplots(1,1,figsize=(6,6))
    cont = axes.contourf(xarray,yarray,zarray,80,cmap = cmap,alpha = alpha)

    plt.xlabel("x")
    plt.ylabel("y")

    for c in cont.collections:
        c.set_edgecolor("face")

    _printout(fname)
    plt.savefig(fname, dpi=25)

def wigner_3d(state,xrange,yrange,N = 100,fname='fig_husimi_3d.eps',
            cmap = 'viridis',alpha = 1.0):
    """
    to visualize a 3d Wigner function

    Parameters:
    ------------
    state: quantum object
        A given quantum state needed to visualize
    xrange, yrange: array-like(2)
        The minimum and maximum values of the coordinates 
    N: integer
        number of steps for xrange, yrange
    fname: string
        File name  
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 

    Returns:
    A file with fname
    """

    xarray = np.linspace(xrange[0],xrange[1],N)
    yarray = np.linspace(yrange[0],yrange[1],N)
    zarray = wigner(state,xarray,yarray)
    zarray /= amax(zarray)

    xx, yy = meshgrid(xarray,yarray)

    norm = plt.Normalize(zarray.min(), zarray.max())
    colors = cm.viridis(norm(zarray))

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.plot_surface(xx, yy, zarray, cmap=cmap,
                    rstride=1, cstride=1,linewidth=0,facecolors=colors)

    plt.xlabel("x")
    plt.ylabel("y")

    _printout(fname)
    plt.savefig(fname, dpi=25)

try:
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D
except:
    pass

def husimi_spin_3d(state,theta,phi,N = 100,cmap = 'viridis',dirname ="",
                fname = 'fig_husimi_spin_3d.eps',alpha = 1,view=(120,120),use_axis=False):
    """ to plot Husimi visualization in Bloch sphere
    
    Parameters:
    state: quantum object
        A given quantum state needed to visualize
    theta, phi: array-like(2)
        The minimum and maximum values of the coordinates 
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 
    fname: string
        File name  
        
    Returns:
    A file with fname
    """

    theta_array = np.linspace(theta[0], theta[1], N)
    phi_array = np.linspace(phi[0], phi[1], N)
    theta_grid, phi_grid = np.meshgrid(theta_array,phi_array)

    #convert to x,y,z
    x = np.sin(theta_grid) * np.cos(phi_grid)
    y = np.sin(theta_grid) * np.sin(phi_grid)
    z = np.cos(theta_grid)

    h = husimi_spin(state,theta_array,phi_array)

    a = str('cm.')+cmap
    if h.min() < -10e10:
        cmap = eval(a)
        norm = mpl.colors.Normalize(-h.max(), h.max())
    else:
        cmap = eval(a)
        norm = mpl.colors.Normalize(h.min(), h.max())

    fig = plt.figure(figsize=(6,6))
    ax = fig.add_subplot(projection='3d')
    ax.plot_surface(x, y, z, rstride=1, cstride=1, shade=False,
                    facecolors=cmap(norm(h)),linewidth=0,alpha=alpha)
    if use_axis:
        ax.quiver(1,0,0,1,0,0,color = 'navy', alpha = .8, lw = 3) #x arrow
        ax.text(1.6,0,0.1,"Sx","x")
        ax.quiver(0,1,0,0,1,0,color = 'navy', alpha = .8, lw = 3)#y arrow
        ax.text(0,1.6,0.1,"Sy","y")
        ax.quiver(0,0,1,0,0,1,color = 'navy', alpha = .8, lw = 3)#z arrow
        ax.text(0.05,0.05,1.6,"Sz","z")
    plt.axis('off')
    _printout(fname)
    ax.view_init(view)
    elev,azim = view
    ax.view_init(elev=elev, azim=azim)
    fig.savefig(os.path.join(dirname,fname),dpi=50,bbox_inches='tight')
    plt.close()

def wigner_spin_3d(state,theta,phi,N = 100,cmap = 'viridis',
                fname = 'fig_wigner_spin_3d.eps',alpha = 1,view=(120,120),use_axis=False):
    """ to plot Husimi visualization in Bloch sphere
    
    Parameters:
    state: quantum object
        A given quantum state needed to visualize
    theta, phi: array-like(2)
        The minimum and maximum values of the coordinates 
    cmap: str or Colormap
        A colormap instance or colormap name (default: 'viridis') 
    fname: string
        File name  
        
    Returns:
    A file with fname
    """

    theta_array = np.linspace(theta[0], theta[1], N)
    phi_array = np.linspace(phi[0], phi[1], N)
    theta_grid, phi_grid = np.meshgrid(theta_array,phi_array)

    #convert to x,y,z
    x = np.sin(theta_grid) * np.cos(phi_grid)
    y = np.sin(theta_grid) * np.sin(phi_grid)
    z = np.cos(theta_grid)

    h = wigner_spin(state,theta_array,phi_array)
    #print(h)

    a = str('cm.')+cmap
    if h.min() < -10e10:
        cmap = eval(a)
        norm = mpl.colors.Normalize(-h.max(), h.max())
    else:
        cmap = eval(a)
        norm = mpl.colors.Normalize(h.min(), h.max())

    fig = plt.figure(figsize=(6,6))
#    ax = fig.gca(projection='3d')
    ax = fig.add_subplot(1,1,1, projection='3d')
    ax.plot_surface(x, y, z, rstride=1, cstride=1, shade=False,
                    facecolors=cmap(norm(h)),linewidth=0,alpha=alpha)
    if use_axis:
        ax.quiver(1,0,0,1,0,0,color = 'navy', alpha = .8, lw = 3) #x arrow
        ax.text(1.6,0,0.1,"Sx","x")
        ax.quiver(0,1,0,0,1,0,color = 'navy', alpha = .8, lw = 3)#y arrow
        ax.text(0,1.6,0.1,"Sy","y")
        ax.quiver(0,0,1,0,0,1,color = 'navy', alpha = .8, lw = 3)#z arrow
        ax.text(0.05,0.05,1.6,"Sz","z")
    plt.axis('off')
    _printout(fname)
    ax.view_init(view)
    elev,azim = view
    ax.view_init(elev=elev, azim=azim)
    fig.savefig(f"{elev},{azim},{fname}",dpi=50,bbox_inches='tight')
    plt.show(block=True)
###
def _printout(fname):
    print('***')
    print('Figure ', fname, ' has created')

def cmindex(d):
    #https://matplotlib.org/3.2.1/tutorials/colors/colormaps.html
    cmaps = ['viridis', #1
            'plasma',  #2
            'inferno', #3
            'magma',   #4 
            'cividis', #5
            'Greys',   #6
            'Purples', #7
            'Blues',   #8
            'Greens',  #9
            'Oranges', #10
            'Reds',    #11
            'YlOrBr',  #12
            'YlOrRd',  #13
            'OrRd',    #14
            'PuRd',    #15
            'RdPu',    #16
            'BuPu',    #17
            'GnBu',    #18
            'PuBu',    #19
            'YlGnBu',  #20
            'PuBuGn',  #21
            'BuGn',    #22
            'YlGn',    #23
            'binary',  #24
            'gist_yarg', #25
            'gist_gray', #26
            'gray',    #27
            'bone',    #28
            'pink',    #29
            'spring',  #30
            'summer',  #31
            'autumn',  #32
            'winter',  #33
            'cool',    #34
            'Wistia',  #35
            'hot',     #36
            'afmhot',  #37
            'gist_heat',#38
            'copper',   #39
            'PiYG',     #40
            'PRGn',     #41
            'BrBG',     #42
            'PuOr',     #43
            'RdGy',     #44
            'RdBu',     #45
            'RdYlBu',   #46
            'RdYlGn',   #47
            'Spectral', #48
            'coolwarm', #49
            'bwr',      #50
            'seismic',  #51
            'twilight', #52
            'twilight_shifted', #53 
            'hsv',              #54
            'Pastel1',  #55
            'Pastel2',  #56
            'Paired',   #57
            'Accent',   #58
            'Dark2',    #59
            'Set1',     #60
            'Set2',     #61
            'Set3',     #62
            'tab10',    #63
            'tab20',    #64
            'tab20b',   #65
            'tab20c',   #66
            'flag',     #67
            'prism',    #78
            'ocean',    #69
            'gist_earth',#70
            'terrain',   #71
            'gist_stern',#72
            'gnuplot',   #73
            'gnuplot2',  #74
            'CMRmap',    #75
            'cubehelix', #76
            'brg',       #77
            'gist_rainbow', #78
            'rainbow',      #79
            'jet',          #80
            'nipy_spectral',#81
            'gist_ncar'     #82
            ]
    if d > len(cmaps):
        raise IndexError('Out of index range')
    else:
        return cmaps[d]



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['hinfo']

import os
import sys
import multiprocessing
import numpy as np

def _mac_hardware_info():
    info = dict()
    results = dict()
    for l in [l.split(':') for l in os.popen('sysctl hw').readlines()[1:]]:
        info[l[0].strip(' "').replace(' ', '_').lower().strip('hw.')] = \
            l[1].strip('.\n ')
    results.update({'cpus': int(info['physicalcpu'])})
    results.update({'cpu_freq': int(float(os.popen('sysctl -n machdep.cpu.brand_string')
                                .readlines()[0].split('@')[1][:-4])*1000)})
    results.update({'memsize': int(int(info['memsize']) / (1024 ** 2))})
    # add OS information
    results.update({'os': 'Mac OSX'})
    return results


def _linux_hardware_info():
    results = {}
    # get cpu number
    sockets = 0
    cores_per_socket = 0
    frequency = 0.0
    for l in [l.split(':') for l in open("/proc/cpuinfo").readlines()]:
        if (l[0].strip() == "physical id"):
            sockets = np.maximum(sockets,int(l[1].strip())+1)
        if (l[0].strip() == "cpu cores"):
            cores_per_socket = int(l[1].strip())
        if (l[0].strip() == "cpu MHz"):
            frequency = float(l[1].strip()) / 1000.
    results.update({'cpus': sockets * cores_per_socket})
    # get cpu frequency directly (bypasses freq scaling)
    try:
        file = "/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq"
        line = open(file).readlines()[0]
        frequency = float(line.strip('\n')) / 1000000.
    except:
        pass
    results.update({'cpu_freq': frequency})

    # get total amount of memory
    mem_info = dict()
    for l in [l.split(':') for l in open("/proc/meminfo").readlines()]:
        mem_info[l[0]] = l[1].strip('.\n ').strip('kB')
    results.update({'memsize': int(mem_info['MemTotal']) / 1024})
    # add OS information
    results.update({'os': 'Linux'})
    return results

def _freebsd_hardware_info():
    results = {}
    results.update({'cpus': int(os.popen('sysctl -n hw.ncpu').readlines()[0])})
    results.update({'cpu_freq': int(os.popen('sysctl -n dev.cpu.0.freq').readlines()[0])})
    results.update({'memsize': int(os.popen('sysctl -n hw.realmem').readlines()[0]) / 1024})
    results.update({'os': 'FreeBSD'})
    return results

def _win_hardware_info():
    try:
        from comtypes.client import CoGetObject
        winmgmts_root = CoGetObject("winmgmts:root\cimv2")
        cpus = winmgmts_root.ExecQuery("Select * from Win32_Processor")
        ncpus = 0
        for cpu in cpus:
            ncpus += int(cpu.Properties_['NumberOfCores'].Value)
    except:
        ncpus = int(multiprocessing.cpu_count())
    return {'os': 'Windows', 'cpus': ncpus}


def hinfo():
    """
    Returns basic hardware information about the computer.

    Gives actual number of CPU's in the machine, even when hyperthreading is
    turned on.

    Returns
    -------
    info : dict
        Dictionary containing cpu and memory information.

    """
    if sys.platform == 'darwin':
        out = _mac_hardware_info()
    elif sys.platform == 'win32':
        out = _win_hardware_info()
    elif sys.platform in ['linux', 'linux2']:
        out = _linux_hardware_info()
    elif sys.platform.startswith('freebsd'):
        out = _freebsd_hardware_info()
    else:
        out = {}
    return out

if __name__ == '__main__':
    print(hinfo())



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

__all__ = ['husimi', 'husimi_spin', 'wigner', 'wigner_spin']

import numpy as np
from scipy.special import genlaguerre,factorial,sph_harm
from tqix import *

def husimi(state,x_array,y_array):
    """ to calculate Husimi Q Function of a quantum state
    by this formula
    Q(a) = 1/pi * <a|rho|a>
    where |a> is a coherent state, a is complex
    rho is density-matrix form of the input state

    Parameters
    ----------
    state: a quantum state
    x_array: an array of coordinate x
    y_array: an array of coordinate y

    ----------
    Return
    Array of Husimi Q function over array [x_array,y_array]

    """
    if not isoperx(state):
        state = operx(state)
    d = shapex(state)[0] #dimension of state
    M = np.shape(x_array)[0]
    N = np.shape(y_array)[0]

    xx,yy = np.meshgrid(x_array,y_array)
    alpha = (xx + 1j*yy)/(np.sqrt(2))
    value = np.zeros((M,N),dtype=float)

    for i in range(M):
        for j in range(N):
            temp = coherent(d,alpha[i,j])
            qfunc = dotx(daggx(temp),state,temp)/(np.pi)
            value[i,j] = float(np.real(qfunc))

    return value

def husimi_spin(state,x_array,y_array):
    """ to calculate Husimi Q Function of a quantum state
    by this formula
    Q(a) = 1/pi * <a|rho|a>
    where |a> is a coherent state, a is complex
    rho is density-matrix form of the input state

    Parameters
    ----------
    state: a quantum state
    x_array: an array of coordinate x
    y_array: an array of coordinate y

    ----------
    Return
    Array of Husimi Q function over array [x_array,y_array]

    """
    if not isoperx(state):
        state = operx(state)
    d = shapex(state)[0] #dimension of state
    M = np.shape(x_array)[0]
    N = np.shape(y_array)[0]

    xx,yy = np.meshgrid(x_array,y_array)
    alpha = (xx + 1j*yy)
    value = np.zeros((M,N),dtype=float)

    for i in range(M):
        for j in range(N):
            temp = spin_coherent(float((d-1)/2.0),np.real(alpha[i,j]),np.imag(alpha[i,j]))
            #temp = coherent(d,alpha[i,j])
            qfunc = dotx(daggx(temp),state,temp) /(np.pi)
            value[i,j] = float(np.real(qfunc))

    return value

def wigner(state,x_array,y_array):
    """ to calculate Wigner Function of a quantum state,
    see formulae (5.105) and (5.108) in 
    Measuring the Quantum State of Light (CUP,1997)
    by Ulf Leonhardt

    Parameters
    ----------
    state: a quantum state
    x_array: an array of coordinate x
    y_array: an array of coordinate y

    ----------
    Return
    Array of Wigner function over array [x_array,y_array]

    """
    if not isoperx(state):
        state = operx(state)
    d = shapex(state)[0] #dimension of state
    M = np.shape(x_array)[0]
    N = np.shape(y_array)[0]

    xx,yy = np.meshgrid(x_array,y_array)
    alpha = (xx + 1j*yy)/(np.sqrt(2))
    value = np.zeros((M,N),dtype=float)

    for i in range(M):
        for j in range(N):
            value[i,j] = float(np.real(_wigner(d,state,alpha[i,j])))
    return value

def _wigner(d,state,alpha):
    """to calculate Wigner Function of a quantum state,
    see formulae (5.105) and (5.108) in 
    Measuring the Quantum State of Light (CUP,1997)
    by Ulf Leonhardt
    """
    result = 0.0
    for i in range(d):
        for j in range(d):
            if (j > i):
                laguerre = genlaguerre(i,j-i)(4 * abs(alpha)**2)
                temp = (-1)**j / (np.pi) * np.sqrt(factorial(i)/factorial(j))\
                        *np.exp(-2 * abs(alpha)**2) * (-2*conjx(alpha))**(j-i)\
                        *laguerre
                result += temp * state[i,j]
            else:
                laguerre = genlaguerre(j,i-j)(4 * abs(alpha)**2)
                temp = (-1)**j / (np.pi) * np.sqrt(factorial(j)/factorial(i))\
                        *np.exp(-2 * abs(alpha)**2) * (-2*conjx(-alpha))**(i-j)\
                        *laguerre
                result += conjx(temp) * state[i,j]
    return result

def clebsch_coe(j1, j2, j3, m1, m2, m3):

    if m3 != m1 + m2:
        return 0
        
    minval = int(np.max([-j1+j2+m3,-j1+m1,0]))
    maxval = int(np.min([j2+j3+m1,j3-j1+j2,j3+m3]))

    CC = np.sqrt((2.0*j3 + 1.0) * factorial(j3 + j1 - j2) *
            factorial(j3 - j1 + j2) * factorial(j1 + j2 - j3) *
            factorial(j3 + m3) * factorial(j3 - m3) /
            (factorial(j1 + j2 + j3 + 1) *
            factorial(j1 - m1) * factorial(j1 + m1) *
            factorial(j2 - m2) * factorial(j2 + m2)))
    SC = 0
    for v in range(minval, maxval + 1):
        SC += (-1.0) ** (v + j2 + m2) / factorial(v) * \
           factorial(j2 + j3 + m1 - v) * factorial(j1 - m1 + v) / \
           factorial(j3 - j1 + j2 - v) / factorial(j3 + m3 - v) / \
           factorial(v + j1 - j2 - m3)
    return CC * SC

def _rho_kq(rho, j, k, q):
    # state in kq basis
    if not isoperx(rho):
        state = operx(rho)
    d = shapex(rho)[0] #dimension of state
    
    rkq = 0j
    jrange = np.linspace(-j,j,d)
    for m1 in jrange: #range(-j, j+1):
        for m2 in jrange: #range(-j, j+1):
            rkq += (-1)**(j - m1 - q) * clebsch_coe(j, j, k, m1, -m2,q) * dotx(daggx(zbasis(j,m1)),rho,zbasis(j,m2))
    return rkq


def wigner_spin(state,x_array,y_array):
    """ to calculate spin Wigner Function of a quantum state,
    see 
    Parameters
    ----------
    state: a quantum state
    x_array: an array of coordinate x
    y_array: an array of coordinate y

    ----------
    Return
    Array of spin Wigner function over array [x_array,y_array]

    """
    if not isoperx(state):
        state = operx(state)
    d = shapex(state)[0] #dimension of state
    j = (d-1)/2
    
    M = np.shape(x_array)[0]
    N = np.shape(y_array)[0]

    xx,yy = np.meshgrid(x_array,y_array)
    alpha = (xx + 1j*yy)

    qfunc = np.zeros_like(xx, dtype=complex)
    for k in range(0,int(2 * j)+1):
         for q in range(-k,k+1):
             qfunc += _rho_kq(state, j, k, q) * sph_harm(q,k,yy,xx)

    return np.real(qfunc)



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

from tqix.dsm.execute import *
from tqix.dsm.dsmStrong import *
from tqix.dsm.dsmWeak import *
from tqix.dsm.dsmProb import *
from tqix.dsm.dsmProb_Conf_1 import *
from tqix.dsm.dsmProb_Conf_2 import *
from tqix.dsm.util import *



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

from tqix.povm.pauli import *
from tqix.povm.stoke import *
from tqix.povm.mub import *
from tqix.povm.sic import *

__all__ = ['_pauli','_stoke','_mub','_sic']

def _pauli(n):
    # to return Pauli POVM of
    # n qubits
    return _pauli_(n)

def _stoke(n):
    # to return Stoke POVM of
    # n qubits
    return _stoke_(n)

def _mub(d):
    # to return MUB POVM of
    # d dimension
    return _mub_(d)

def _sic(d):
    # to return SIC POVM of
    # d dimension
    return _sic_(d)




"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

from numpy import sin,pi,tan,sqrt

import numpy as np
import matplotlib.pyplot as plt

from tqix.qobj import *
from tqix.utility import randunit
from tqix.backend import cdf
from tqix.dsm.util import gtrace,gfide

__all__ = ['execu']

def execu(state,niter,nsamp,theta,pHist='False'):
   # to run weak DSM
    
   dim  = state.shape[0]
   prob = get_prob(state,theta,dim)

   ntr = np.zeros(nsamp)
   nfi = np.zeros(nsamp)
   
   if typex(state)=='oper':
      for i in range(nsamp):
         rho10,rho11 = get_rho10_rho11(state,prob,niter,dim)
         restate = get_qurec(state,theta,rho10,rho11,dim)
         ntr[i] = gtrace(restate,state)
         nfi[i] = gfide(restate,state)
   else:
      for i in range(nsamp):
         restate = get_pqurec(state,theta,niter,prob,dim)
         ntr[i] = gtrace(restate,state)
         nfi[i] = gfide(restate,state)

   # if plot hist
   if pHist:
      plt.hist(nfi,bins=100)
      #plt.show()

   # take average
   atr = sum(ntr)/float(nsamp)
   afi = sum(nfi)/float(nsamp)
   etr,efi = 0.0,0.0

   for i in range(nsamp):
      etr += (ntr[i]-atr)**2
      efi += (nfi[i]-afi)**2

   etr /= float(nsamp)
   efi /= float(nsamp)

   return atr,np.sqrt(etr),afi,np.sqrt(efi)

"""
-------------------
some common modules 
-------------------
"""

# get propability
def get_prob(state,theta,dim):
   # to get probability

   sin_tt  = sin(pi*theta)
   sin_2tt = sin(pi*theta/2.0)**2
   epsi_tt = 2.0*sin_2tt

   if typex(state)=='oper':
      temp_prob = np.zeros((4,dim,dim),dtype=complex)
      prob = np.zeros((5,dim,dim),dtype=complex)
      for x in range(dim):
         for p in range(dim):
            # temp_prob_00
            #temp_prob[0,x,p] = 0.0
            for tx in range(dim):
               for y in range(dim):
                  temp_prob[0,x,p] += \
                  state[tx,y]*np.exp(1j*2.0*pi*(y-tx)*p/dim)
            for y in range(dim):
               temp_prob[0,x,p] -= 2.0*sin_2tt*(\
               state[x,y]*np.exp(1j*2.0*pi*(y-x)*p/dim)+\
               np.conj(state[x,y])*np.exp(-1j*2.0*pi*(y-x)*p/dim))

            temp_prob[0,x,p] += 4*sin_2tt**2*state[x,x]
            temp_prob[0,x,p] /= dim

            # temp_prob_10
            for y in range(dim):
               temp_prob[1,x,p] += state[x,y]*np.exp(1j*2.0*pi*(y-x)*p/dim)
            temp_prob[1,x,p] -= 2*sin_2tt*state[x,x]
            temp_prob[1,x,p] = temp_prob[1,x,p]*sin_tt/dim
            
            # temp_prob_01
            temp_prob[2,x,p] = np.conj(temp_prob[1,x,p])

            # temp_prob_11
            temp_prob[3,x,p] = sin_tt**2*state[x,x]/dim

      prob[0,:,:] = np.real((temp_prob[0,:,:]+temp_prob[1,:,:]\
                 +temp_prob[2,:,:]+temp_prob[3,:,:])/2.0) #Prob.+
      prob[1,:,:] = (temp_prob[0,:,:]-temp_prob[1,:,:]\
                  -temp_prob[2,:,:]+temp_prob[3,:,:])/2.0 #Prob.-
      prob[2,:,:] = (temp_prob[0,:,:]-1j*temp_prob[1,:,:]\
                  +1j*temp_prob[2,:,:]+temp_prob[3,:,:])/2.0 #Prob.L
      prob[3,:,:] = (temp_prob[0,:,:]+1j*temp_prob[1,:,:]\
                  -1j*temp_prob[2,:,:]+temp_prob[3,:,:])/2.0 #Prob.R
      prob[4,:,:] = temp_prob[3,:,:] #Prob.1
      return np.real(prob)
   else: #for pure state
      rpsi = np.zeros((dim))
      ipsi = np.zeros((dim))
      prob = np.zeros((5,dim))
        
      rpsi[:] = np.real(state[:,0])
      ipsi[:] = np.imag(state[:,0])
      psit = sqrt(sum(rpsi)**2+sum(ipsi)**2)

      prob[0,:]= psit/dim*(psit/2.0+pi*theta*rpsi[:]) #P(+)
      prob[1,:]= psit/dim*(psit/2.0-pi*theta*rpsi[:]) #P(-)
      prob[2,:]= psit/dim*(psit/2.0+pi*theta*ipsi[:]) #P(L)
      prob[3,:]= psit/dim*(psit/2.0-pi*theta*ipsi[:]) #P(R)
      prob[4,:]= 0.0 #P(1)
      return np.real(prob)

def get_bisection(state,prob,dim):
   # to calculate bisection

   if typex(state)=='oper':
      rs = np.zeros((5,dim,dim))
      for i in range(5):
         for j in range(dim):
            for k in range(dim):
               rs[i,j,k] = randunit()/prob[i,j,k]
   else:
      rs = np.zeros((5,dim))
      for i in range(5):
         for j in range(dim):
            rs[i,j] = randunit()/prob[i,j]
   return rs

def get_rho10_rho11(state,prob,niter,dim):
   # to calculate rho10 and rho11
   # just for mixed state

   N = np.int(niter/2)
   rho10 = np.zeros((dim,dim),dtype=complex)
   rho11 = np.zeros((dim,dim),dtype=complex)
   temp_rs = np.zeros((N,5,dim,dim))

   for i in range(N):
      temp_rs[i,:,:,:] = get_bisection(state,prob,dim)

   ave_rs = np.zeros((5,dim,dim))
   for i in range(5):
      for j in range(dim):
         for k in range(dim):
            ave_rs[i,j,k] = cdf(temp_rs[:,i,j,k])
   for x in range(dim):
      for p in range(dim):
         rho10[x,p] = 1/2.0*(ave_rs[0,x,p]-ave_rs[1,x,p]+\
                         1j*(ave_rs[2,x,p]-ave_rs[3,x,p]))
         rho11[x,p] = ave_rs[4,x,p]
   return rho10,rho11

def get_qurec(state,theta,rho10,rho11,dim):
   # to calculate the reconstructed state
   # just for mixed state
    
   restate = np.zeros((dim,dim),dtype=complex)
   for x in range(dim):
      for y in range(dim):
         for p in range(dim):
            restate[x,y] += np.exp(1j*2*pi*(x-y)*p/dim)*\
                            rho10[x,p] #S37
   
   restate = normx(restate)
   return restate

def get_pqurec(state,theta,niter,prob,dim):
   # to get pure quantum state
  
   sin_tt  = sin(pi*theta)
   rpsi = np.real(state)
   ipsi = np.imag(state)
   psit = sqrt(sum(rpsi)**2+sum(ipsi)**2)
   p0 = (sum(rpsi)**2+sum(ipsi)**2)/dim

   N = np.int(niter*p0/2)
   temp_rs = np.zeros((N,5,dim))

   for i in range(N):
      temp_rs[i,:,:] = get_bisection(state,prob,dim)

   ave_rs = np.zeros((5,dim))  
   for i in range(5):
      for j in range(dim):
         ave_rs[i,j] = cdf(temp_rs[:,i,j])

   rtemp = np.zeros(dim)
   itemp = np.zeros(dim)
   rnorm = 0.0
   for i in range(dim):
      rtemp[i] = (ave_rs[0,i]-ave_rs[1,i]+2.0*tan(pi*theta/2.0)*\
             ave_rs[4,i])*dim/(2*psit*sin_tt)
      itemp[i] = (ave_rs[2,i]-ave_rs[3,i])*dim/(2*psit*sin_tt)
      rnorm += rtemp[i]**2+itemp[i]**2
   rtemp /= sqrt(rnorm)
   itemp /= sqrt(rnorm)

   restate = np.zeros((dim,dim),dtype=complex)
   for i in range(dim):
      for j in range(dim):
         restate[i,j] = (rtemp[i]+1j*itemp[i])*\
                     (rtemp[j]-1j*itemp[j])
   return restate



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Viet NT
>>> all rights reserved
________________________________
"""

#from numpy import
from functools import partial
import numpy as np
import random
from tqix.qobj import *
from tqix.qtool import dotx
from tqix.pis.util import *
from tqix.pis import *
from scipy.sparse import block_diag,csr_matrix
from scipy.sparse.linalg import expm
from functools import *
import torch 

__all__ = ['Gates']

class Gates(object):
    """
    Class contains gate operations of tqix.pis
    """       

    def __init__(self):
        """init class of Gates 
        """        
        self.state = None        
        self.theta = None

    def RX(self,theta=None,*args, **kwargs):
        """

        :math:`R_x(\\theta) = e^{-i\\theta J_x}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RX  
        :rtype: ndarray, tensor, sparse
        """        
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("Rx",*args, **kwargs)
    
    def RY(self,theta=None,*args, **kwargs):
        """

        :math:`R_y(\\theta) = e^{-i\\theta J_y}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RY
        :rtype: ndarray, tensor, sparse
        """       
        params = {"theta":theta}
        self.check_input_param(params) 
        return self.gates("Ry",*args, **kwargs)
    
    def RZ(self,theta=None,*args, **kwargs):
        """

        :math:`R_z(\\theta) = e^{-i\\theta J_z}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RZ  
        :rtype: ndarray, tensor, sparse
        """        
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("Rz",*args, **kwargs)
    
    def OAT(self,theta,gate_type,*args, **kwargs):
        """

        :math:`U_{OAT} = e^{-it\\chi J_\\alpha^2}`

        :param theta: angle
        :type theta: float
        :param gate_type: type of gate
        :type gate_type: str
        :return: new state after being acted upon by OAT
        :rtype: ndarray, tensor, sparse
        """

        params = {"theta":theta,"gate_type":gate_type}
        self.check_input_param(params)
        gate_type = gate_type.lower()
        if gate_type =="x":
            return self.RX2(theta,*args, **kwargs)
        if gate_type =="y":
            return self.RY2(theta,*args, **kwargs)
        if gate_type =="z":
            return self.RZ2(theta,*args, **kwargs)                    
    
    def TAT(self,theta,gate_type,*args, **kwargs):
        """

        :math:`U_{TAT} = e^{-it\\chi (J_\\alpha^2 - J_\\beta^2)}`

        :param theta: angle 
        :type theta: float
        :param gate_type: type of gate
        :type gate_type: str
        :return: new state after being acted upon by TAT 
        :rtype: ndarray, tensor, sparse
        """             
        params = {"theta":theta,"gate_type":gate_type}
        self.check_input_param(params)
        return self.gates(type=gate_type+"TAT",*args, **kwargs)
    
    def TNT(self,theta,gate_type,omega,*args, **kwargs):
        """

        :math:`U_{TNT} = e^{-it(\\chi J_\\alpha^2-\\Omega J_\\beta)}`

        :param theta: angle
        :type theta: float
        :param gate_type: type of gate
        :type gate_type: str
        :return: new state after being acted upon by TNT
        :rtype: ndarray, tensor, sparse
        """        
        params = {"theta":theta,"gate_type":gate_type,"omega":omega}
        self.check_input_param(params)
        return self.gates(type=gate_type+"TNT",omega=omega,*args, **kwargs)

    def RX2(self,theta=None,*args, **kwargs):
        """

        :math:`R_{x^2}(\\theta) = e^{-i\\theta J_x^2}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RX2
        :rtype: ndarray, tensor, sparse
        """        
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("Rx2",*args, **kwargs)
    
    def RY2(self,theta=None,*args, **kwargs):
        """

        :math:`R_{y^2}(\\theta) = e^{-i\\theta J_y^2}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RY2
        :rtype: ndarray, tensor, sparse
        """  
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("Ry2",*args, **kwargs)
    
    def RZ2(self,theta=None,*args, **kwargs):
        """

        :math:`R_{z^2}(\\theta) = e^{-i\\theta J_z^2}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by RZ2
        :rtype: ndarray, tensor, sparse
        """   
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("Rz2",*args, **kwargs)
    
    def R_plus(self,theta=None,*args, **kwargs):
        """

        :math:`R_+(\\theta) = e^{-i\\theta J_+}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by R_plus
        :rtype: ndarray, tensor, sparse
        """  
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("R+",*args, **kwargs)
    
    def R_minus(self,theta=None,*args, **kwargs):
        """

        :math:`R_-(\\theta) = e^{-i\\theta J_-}`

        :param theta: angle, defaults to None
        :type theta: float, optional
        :return: new state after being acted upon by R_minus
        :rtype: ndarray, tensor, sparse
        """  
        params = {"theta":theta}
        self.check_input_param(params)
        return self.gates("R-",*args, **kwargs)
    
    def GMS(self,theta,phi,*args, **kwargs):
        """

        :math:`U_{GMS} = e^{-i\\theta (J_x\\cos\\phi + J_y\\sin\\phi)^2}`

        :param theta: theta angle
        :type theta: float
        :param phi: phi angle
        :type phi: float
        :return: new state after being acted upon by GMS
        :rtype: ndarray, tensor, sparse
        """        

        params = {"theta":theta,"phi":phi}
        self.check_input_param(params)
        return self.gates(type="GMS",phi=phi,*args, **kwargs)

    def RN(self,theta,phi,*args, **kwargs):
        """
        
        :math:`R_{n}(\\theta,\\phi)= e^{-i\\theta J_{n}}={exp}[i\\theta (J_x\\sin\\phi - J_y\\cos\\phi)]`

        :param theta: theta angle
        :type theta: float
        :param phi: phi angle
        :type phi: float
        :return: new state after being acted upon by RN
        :rtype: ndarray, tensor, sparse
        """              
        params = {"theta":theta,"phi":phi}
        self.check_input_param(params)
        return self.gates(type="RN",phi=phi,*args, **kwargs)

    def R_phase(self,theta,phi,gamma,*args, **kwargs):
        """
        
        :math:`R_{phases}(\\phases)= e^{-i\\phases J_{n}}`

        :param theta: theta angle
        :type theta: float
        :param phi: phi angle
        :type phi: float
        :param gamma: gamma angle
        :type gamma: float
        :return: new state after being acted upon by R_phases
        :rtype: ndarray, tensor, sparse
        """              
        params = {"theta":theta,"phi":phi, "gamma":gamma}
        self.check_input_param(params)
        return self.gates(type="RP",phi=phi,gamma=gamma,*args, **kwargs)

    def check_input_param(self,params):
        """

        :param params: parameters to check
        :type params: dict
        :raises ValueError: the parameter is None
        """        
        
        self.theta = params["theta"]
        for param,value in params.items():
            if value == None:
                raise ValueError(f"{param} is None")
    
    def get_N_d_d_dicked(self,state):
        """

        :param state: state of qubits
        :type state: ndarray, tensor, sparse
        :return: d_in
        :rtype: int
        :return: N_in
        :rtype: int
        :return: d_dicke
        :rtype: int
        """        

        d_in = shapex(state)[0]
        N_in = self.N
        d_dicke = get_dim(N_in)
        return d_in,N_in,d_dicke

    def get_J(self,N_in,d_in,d_dicke,type):
        """

        :param N_in: number of qubits
        :type N_in: int
        :param d_in: dimension of state
        :type d_in: int
        :param d_dicke: dimension of dicke basis 
        :type d_dicke: int
        :param type: type of J operator
        :type type: str
        :return: operator matrix representation 
        :rtype: ndarray, tensor, sparse
        """                           
        
        if "x" in type:
            S = partial(Sx)
        
        elif "y" in type:
            S = partial(Sy)
        
        elif "z" in type:
            S = partial(Sz)
        
        elif "+" in type:
            S = partial(S_plus)
        
        elif "-" in type:
            S = partial(S_minus)

        if d_in != d_dicke:
            if "2" in type:
                if self.use_gpu:
                    J = S(N_in/2,self.use_gpu,self.device).mm(S(N_in/2,self.use_gpu,self.device))
                else:
                    J = S(N_in/2).dot(S(N_in/2))
            else:
                J = S(N_in/2,self.use_gpu,self.device)
        else:
            j_array = get_jarray(N_in)[::-1]
            blocks = []
            for j in j_array:
                if self.use_gpu:
                    blocks.append(S(j,self.use_gpu,self.device))
                else:
                    blocks.append(S(j))
            if not self.use_gpu:
                J = block_diag(blocks,format="csc")
            else:
                J = torch.block_diag(*blocks)
            if "2" in type:
                if self.use_gpu:
                    J = J.mm(J)
                else:
                    J = J.dot(J)     
        return J
    
    def Jx(self):
        """

        :return: Jx
        :rtype: ndarray, tensor, sparse
        """            
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(self.state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        return get_J("x")
    
    def Jy(self):
        """

        :return: Jy
        :rtype: ndarray, tensor, sparse
        """      
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(self.state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        return get_J("y")
    
    def Jz(self):
        """

        :return: Jz
        :rtype: ndarray, tensor, sparse
        """      
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(self.state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        return get_J("z")
    
    def J_plus(self):
        """

        :return: J_plus
        :rtype: ndarray, tensor, sparse
        """      
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(self.state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        return get_J("+")
    
    def J_minus(self):
        """

        :return: J_minus
        :rtype: ndarray, tensor, sparse
        """    
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(self.state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        return get_J("-")
    
    def var(self,type="",*args, **kwargs):
        """

        :param type: type of variance, defaults to ""
        :type type: str, optional
        :return: variance of state 
        :rtype: complex
        """            
        use_vector = kwargs.pop('use_vector', None)
        n = kwargs.pop('n', None)
        if use_vector:
            exp_val_J_2 = self.expval(type=type+"2",n=n,use_vector=use_vector)
            exp_val_J = self.expval(type=type,n=n,use_vector=use_vector)
        else:
            if "theta" in type:
                exp_val_J_2 = self.expval(type=type+"2")
                exp_val_J = self.expval(type=type)
            else:
                exp_val_J_2 = self.expval(type=type+"2")
                exp_val_J = self.expval(type=type)
        return exp_val_J_2-exp_val_J**2
    
    def expval(self,type="",*args, **kwargs):
        """

        :param type: type of expectation, defaults to ""
        :type type: str, optional
        :return: expected value of state 
        :rtype: complex
        """          
        state = self.state
        observable = kwargs.pop('observable', None)
        if observable is not None:
            if self.use_gpu:
                return observable.mm(state).diagonal().sum()
            return observable.dot(state).diagonal().sum()
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        type = type.lower()
        count_ops = 0
        for ops in ["x","y","z","+","-"]:
            if ops in type:
                count_ops += 1
        if "j" in type:
                count_ops += 3
        if"cov" in type:
            count_ops += 6
        if "theta" in type or "n1n2_minus" in type or "n1n2_plus" in type:
            count_ops += 2
        
        if "oat'" in type:
            J = get_J(type[4]+'2')
            if '2' in type:
                if self.use_gpu:
                    J = J.mm(J)
                else:
                    J = J.dot(J)
        elif "tnt'" in type:
            J_2 = get_J(type[4]+"2")
            J_prime = get_J(type[5])
            J = (J_2 - J_prime*N_in/2)
            if '2' in type:
                if self.use_gpu:
                    J = J.mm(J)
                else:
                    J = J.dot(J)
        elif "tat'" in type:
            J_2 = get_J(type[4]+"2")
            J_2_prime = get_J(type[5]+"2")
            J = (J_2-J_2_prime)
            if '2' in type:
                if self.use_gpu:
                    J = J.mm(J)
                else:
                    J = J.dot(J)
        else:
            if count_ops == 1:
                J = get_J(type)
            elif count_ops > 1:
                use_vector = kwargs.pop('use_vector', None)
                n = kwargs.pop('n', None)
                if use_vector:
                    order = {"x":0,"y":1,"z":2}

                    if "cov" in type or "n1n2_minus" in type or "n1n2_plus" in type:
                        n1 = kwargs.pop('n1', None)
                        n2 = kwargs.pop('n2', None)
                        if self.use_gpu:
                            list_J_1 = ([get_J(type_J)*(n1[order[type_J]]) for type_J in "xyz"])
                            list_J_2 = ([get_J(type_J)*(n2[order[type_J]]) for type_J in "xyz"])
                        else:
                            list_J_1 = ([get_J(type_J).dot(n1[order[type_J]]) for type_J in "xyz"])
                            list_J_2 = ([get_J(type_J).dot(n2[order[type_J]]) for type_J in "xyz"])
                        J_1 = reduce(lambda x,y:x+y,list_J_1)
                        J_2 = reduce(lambda x,y:x+y,list_J_2)
                        if "cov" in type:
                            if self.use_gpu:
                                J = J_1.mm(J_2) + J_2.mm(J_1)
                            else:
                                J = J_1.dot(J_2) + J_2.dot(J_1)
                        elif "n1n2_minus" in type:
                            if self.use_gpu:
                                J_1 = J_1.mm(J_1)
                                J_2 = J_2.mm(J_2)
                            else:
                                J_1 = J_1.dot(J_1)
                                J_2 = J_2.dot(J_2)
                            J = J_1 - J_2
                        elif "n1n2_plus" in type:
                            if self.use_gpu:
                                J_1 = J_1.mm(J_1)
                                J_2 = J_2.mm(J_2)
                            else:
                                J_1 = J_1.dot(J_1)
                                J_2 = J_2.dot(J_2)
                            J = J_1 + J_2
                    else:
                        if self.use_gpu:
                            list_J = ([get_J(type_J)*(n[order[type_J]]) for type_J in type if type_J != "2"])
                        else:
                            list_J = ([get_J(type_J).dot(n[order[type_J]]) for type_J in type if type_J != "2"])
                        J = reduce(lambda x,y:x+y,list_J)
                        if "2" in type:
                            if self.use_gpu:
                                J = J.mm(J)
                            else:
                                J = J.dot(J)
                elif "theta" in type:
                    J = np.cos(self.theta)*get_J("y")+np.sin(self.theta)*get_J("z")
                    if "2" in type:
                        if self.use_gpu:
                            J = J.mm(J)
                        else:
                            J = J.dot(J)
                else:
                    if "j" in type and "2" in type:
                        list_J = ([get_J(type_J+"2") for type_J in "xyz"])
                    elif "2" in type:
                        list_J = ([get_J(type_J+"2") for type_J in type if type_J != "2"])
                    J = reduce(lambda x,y:x+y,list_J)
        if self.use_gpu:
            return (J.type(torch.complex128) @ state.type(torch.complex128)).diagonal().sum()
        else:
            return J.dot(state).diagonal().sum()


    def gates(self,type="",*args, **kwargs):
        """

        :param type: type of gate, defaults to ""
        :type type: str, optional
        :return: new state after being acted upon by gate operation 
        :rtype: ndarray, tensor, sparse
        """        
        state = self.state
        d_in,N_in,d_dicke = self.get_N_d_d_dicked(state)
        get_J = partial(self.get_J,N_in,d_in,d_dicke)
        
        type = type.lower()
        count_ops = 0
        for ops in ["x","y","z","+","-"]:
            if ops in type:
                count_ops += 1
        if any([True for gate in ["rn","gms","rp"] if gate in type]):
            count_ops += 2
        if count_ops == 1:
            J = get_J(type)
            if self.use_gpu:
                expJ = torch.matrix_exp(-1j*self.theta*J)
            else:
                expJ = expm(-1j*self.theta*J)
        elif count_ops == 2:
            if "tat" in type:
                J_2 = get_J(type[0]+"2")
                J_2_prime = get_J(type[1]+"2")
                if self.use_gpu:
                    expJ = torch.matrix_exp(-1j*self.theta*(J_2-J_2_prime))
                else:
                    expJ = expm(-1j*self.theta*(J_2-J_2_prime))
            elif "tnt" in type:
                J_2 = get_J(type[0]+"2")
                J_prime = get_J(type[1])
                omega = kwargs.pop('omega', None)
                if self.use_gpu:
                    expJ = torch.matrix_exp(-1j*(self.theta*J_2 - omega*J_prime))
                else:
                    expJ = expm(-1j*(self.theta*J_2 - omega*J_prime))
            elif "gms" in type:
                phi = kwargs.pop('phi', None)
                J = get_J("x")
                J_prime = get_J("y")
                S_phi = 2*(J*np.cos(phi)+J_prime*np.sin(phi))
                if self.use_gpu:
                    expJ = torch.matrix_exp(-1j*self.theta*(S_phi@S_phi)/4)
                else:
                    expJ = expm(-1j*self.theta*S_phi.dot(S_phi)/4)
            elif "rn" in type:
                phi = kwargs.pop('phi', None)
                J = get_J("x")
                J_prime = get_J("y")
                if self.use_gpu:
                    expJ = torch.matrix_exp(1j*self.theta*(J*np.sin(phi)-J_prime*np.cos(phi)))
                else:
                    expJ = expm(1j*self.theta*(J*np.sin(phi)-J_prime*np.cos(phi)))
            elif "rp" in type:
                phi = kwargs.pop('phi', None)
                gamma = kwargs.pop('gamma', None)
                J = get_J("x")
                J_prime = get_J("y")
                J_3 = get_J("z")
                if self.use_gpu:
                    expJ = torch.matrix_exp(-1j*(self.theta*J+phi*J_prime+gamma*J_3))
                else:
                    expJ = expm(-1j*(self.theta*J+phi*J_prime+gamma*J_3))

        expJ_conj = daggx(expJ)
        if self.use_gpu:
            new_state = expJ.type(torch.complex128) @ self.state.type(torch.complex128) @ expJ_conj.type(torch.complex128)
        else:
            new_state = expJ.dot(self.state).dot(expJ_conj)
        self.state = new_state

        if args != ():
            noise = args[0]
        else:
            noise = kwargs.pop('noise', None)
        
        if noise is not None:
            if not self.use_gpu:
                if self.num_process is not None:
                    new_state = add_noise(self,noise,num_process=self.num_process)
                else:
                    new_state = add_noise(self,noise)
            else:
                if self.num_process is not None:
                    new_state = add_noise(self,noise,num_process=self.num_process,use_gpu=self.use_gpu,device=self.device)
                else:
                    new_state = add_noise(self,noise,use_gpu=self.use_gpu,device=self.device)
            self.state = new_state

        return self
    
    def measure(self,num_shots = None):
        """

        :param num_shots: number of shots for measurement, defaults to None
        :type num_shots: int, optional
        :return: resulted measurement 
        :rtype: ndarray, tensor, sparse
        """              
        state = self.state 
        device = self.device
        return_tensor = False
        if torch.is_tensor(state):
            if state.is_cuda:
                state = state.detach().cpu().numpy()
            else:
                state = state.numpy()
            return_tensor = True 
        t_prob = np.real(csr_matrix(state).diagonal())
        result = np.zeros_like(t_prob)
        mask_zeros = t_prob != 0
        for _ in range(num_shots):    
            rand_prob = np.array([random.uniform(0,1)]*t_prob.shape[0])
            mask_prob_ge = t_prob > rand_prob
            result[mask_zeros & mask_prob_ge] += 1
        result /= num_shots
        result[-1] = 1 - np.sum(result[:-1])
        if result[-1] < 0:
            result[-1] = 0
        if return_tensor:
            return torch.tensor(result).to(device)
        return result   



from tqix import *
from tqix.qtool import dotx,tensorx
from itertools import product

__all__ = ['_stoke_']

def _stoke_(n):
    """
    Stoke POVM

    Parameters:
    ----------
    n : number of qubits
    
    Return:
    ------
    set of POVM accroding Stoke parameters
    """
    squbit = _stoke_base()
    nqubit = []
    T = []
    T = list(product((squbit),repeat = n))
    for i in range(len(T)):
        M = []
        for j in T[i]:
           M.append(j)
        nqubit.append(tensorx(M))
    return nqubit

def _stoke_base():
    """
    to define four stock parameters
    there are defined to be H, V, D, R
    see: PRA.64.052312

    Return:
    -------
    stoke povm 
    """
    # default
    u = obasis(2,0)
    d = obasis(2,1)
    h = dotx(u,daggx(u))
    v = dotx(d,daggx(d))
    d = dotx(normx(u-d),daggx(normx(u-d)))
    r = dotx(normx(u-1j*d),daggx(normx(u-1j*d)))
    m = [h,v,d,r]
    return m




# -*- coding: utf-8 -*-
""" usefull code for variational quantum entanglement
"""

__all__ = ['training']


import qiskit
import tqix

import numpy as np
import copy
#from autograd.numpy.linalg import inv

def training(qc: qiskit.QuantumCircuit,
            qcirs,
            optimizer,
            num_steps,
            ofset):
            
    """ to run vqent
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: name of circuit function
        - optimizer: optimizer
        - num_steps: number of steps
        - ofset: additional parameters for changing ce

    Returns:
        - params (Numpy array): the optimized parameters
        - cost (Numpy array): the list of loss_value
        - ce: concerntrable entanglement
    """
    
    #model = vqa.circuits.create_ansatz(qc,params,**kwargs)    
    params, costs, ce = fit(qc,qcirs,optimizer,num_steps,ofset)
    
    return [params], costs, ce


def _cost_func(qc,qcirs,ofset):
    
    """Return the cost function
        C = np.abs(ofset - ce)
    
    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: name of circuit function
        - ofset: additional parameters for changing ce

    Returns:
        - Numpy array: The cost function
    """
    #calculate ce 
    fcir = tqix.vqa.vqm.qc_add(qc.copy(), qcirs)
    ces = tqix.vqa.entanglement.concentratable_entanglement_theor(fcir)
    return np.abs(ofset - ces), ces


def _grad_func(qc,qcirs,ofset):
    
    """Return the gradient of the loss function
        C = np.abs(ofset - ces)
        => nabla_C

    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - qcirs: name of circuit function
        - ofset: additional parameters for changing ce

    Returns:
        - Numpy array: The vector of gradient
    """
    
    
    s = tqix.vqa.constants.step_size
    grad_cost = [] #to reset index

    indx = list(range(0, len(qcirs[0][2])))       
    for i in indx:
        qcirs1, qcirs2 = copy.deepcopy(qcirs), copy.deepcopy(qcirs)
        qcirs1[0][2][i] += s #[0] ansatz_name, [2] ansatz_params [0] first_param
        qcirs2[0][2][i] -= s

        cost1,ces1 = _cost_func(qc.copy(),qcirs1,ofset)
        cost2,ces2 = _cost_func(qc.copy(),qcirs2,ofset)
        
        grad_cost.append((cost1 - cost2)/(2*s))
    return grad_cost


def fit(qc: qiskit.QuantumCircuit,qcirs,optimizer,num_steps,ofset):
    
    """Return the new thetas that fit with the circuit from create_circuit_func function#

    Args:
        - qc (QuantumCircuit): The quantum circuit want to calculate the gradient
        - cirs: name of circuit function
        - optimizer: optimizer
        - num_steps: number of steps
        - ofset: additional parameters for changing ce

    Returns:
        - thetas (Numpy array): the optimized parameters
        - loss_values (Numpy array): the list of loss_value
    """
    
    costs = []
    
    for i in range(0, num_steps):
        dev_cost = _grad_func(qc.copy(),qcirs,ofset)
        params = qcirs[0][2]
        
        # update params
        params = optimizer(params,dev_cost,i)
        qcirs[0][2] = params
        
        # compute cost    
        cost,ces = _cost_func(qc.copy(),qcirs,ofset)
        
        costs.append(cost)
    
    return params, costs, ces


def sgd(params, dev_cost, i):
    """optimizer standard gradient descene
    
    Args:
        - params: parameters
        - dev_cost: gradient of the cost function
        - i: use for adam (dont use here)
    
    Returns:
        - params
    """
    learning_rate = tqix.vqa.constants.learning_rate
    params -= learning_rate*dev_cost
    return params


def adam(params, dev_cost, iteration):
    """optimizer adam
    
    Args:
        - params: parameters
        - dev_cost: gradient of the cost function
    
    Returns:
        - params
    """    
    num_params = len(params) #.shape[0]
    beta1, beta2,epsilon = 0.8, 0.999, 10**(-8)
    
    m, v = list(np.zeros(num_params)), list(
                    np.zeros(num_params))
    
    for i in range(num_params):
        m[i] = beta1 * m[i] + (1 - beta1) * dev_cost[i]
        v[i] = beta2 * v[i] + (1 - beta1) * dev_cost[i]**2
        mhat = m[i] / (1 - beta1**(iteration + 1))
        vhat = v[i] / (1 - beta2**(iteration + 1))
        
        params[i] -= tqix.vqa.constants.learning_rate * mhat / (np.sqrt(vhat) + epsilon)
    
    return params
    
    
    


"""The Adam and AMSGRAD optimizers."""

from typing import Optional, Callable, Tuple
import logging
import os
import csv
import numpy as np
from tqix.pis import *
from tqix import * 
from scipy.sparse import block_diag
from scipy import linalg
import torch 
import math 
import time 
logger = logging.getLogger(__name__)

# pylint: disable=invalid-name
__all__ = ['ADAM','GD','gradient_num_diff']

def gradient_num_diff(x_center, f, epsilon, max_evals_grouped=1):
    """

    :param x_center: point around which we compute the gradient
    :type x_center: ndarray
    :param f: the function of which the gradient is to be computed.
    :type f: func
    :param epsilon: the epsilon used in the numeric differentiation.
    :type epsilon: float
    :param max_evals_grouped: max evals grouped, defaults to 1
    :type max_evals_grouped: int, optional
    :return: the gradient computed
    :rtype: ndarray
    """    

    # forig = f(*((x_center,)))

    grad = []
    ei = np.zeros((len(x_center),), float)
    todos_plus = []
    for k in range(len(x_center)):
        ei[k] = 1.0
        d = epsilon * ei
        todos_plus.append(x_center + d)
        ei[k] = 0.0

    todos_minus = []
    for k in range(len(x_center)):
        ei[k] = 1.0
        d = epsilon * ei
        todos_minus.append(x_center - d)
        ei[k] = 0.0

    counter = 0
    chunk_plus = []
    chunks_plus = []
    length = len(todos_plus)
    # split all points to chunks, where each chunk has batch_size points
    for i in range(length):
        x = todos_plus[i]
        chunk_plus.append(x)
        counter += 1
        # the last one does not have to reach batch_size
        if counter == max_evals_grouped or i == length - 1:
            chunks_plus.append(chunk_plus)
            chunk_plus = []
            counter = 0
    
    counter = 0
    chunk_minus = []
    chunks_minus = []
    length = len(todos_minus)
    # split all points to chunks, where each chunk has batch_size points
    for i in range(length):
        x = todos_minus[i]
        chunk_minus.append(x)
        counter += 1
        # the last one does not have to reach batch_size
        if counter == max_evals_grouped or i == length - 1:
            chunks_minus.append(chunk_minus)
            chunk_minus = []
            counter = 0

    for chnk_pl,chnk_mn in zip(chunks_plus,chunks_minus):  # eval the chunks in order
        parallel_parameters_plus = np.concatenate(chnk_pl)
        todos_results_plus = f(parallel_parameters_plus)  # eval the points in a chunk (order preserved)
        parallel_parameters_minus = np.concatenate(chnk_mn)
        todos_results_minus = f(parallel_parameters_minus)
        if isinstance(todos_results_plus, float) and isinstance(todos_results_minus, float):
            grad.append((todos_results_plus - todos_results_minus) / (2*epsilon) )
        else:
            for todor_pl,todor_ms in zip(todos_results_plus,todos_results_minus):
                grad.append((todor_pl - todor_ms) / (2*epsilon) )

    return np.array(grad)

class GD:
    """
        Gradient descent optimizer class (ref. qiskit)
    """    
    def __init__(self,lr: float,eps : float,maxiter: int,tol: float = 1e-6,use_qng=False,route=None,N=None,theta=None):
        """

        :param lr: learning rate
        :type lr: float
        :param eps: Value >=0, Epsilon to be used for finite differences if no analytic gradient method is given.
        :type eps: float
        :param maxiter: maximum iterations
        :type maxiter: int
        :param tol: tolerance for termination, defaults to 1e-6
        :type tol: float, optional
        :param use_qng: option to use quantum gradient descent, defaults to False
        :type use_qng: bool, optional
        :param route: the structure of circuit; for example, if we have a circuit which
        has route=(("RN2",),("OAT","Z"),("TNT","ZX"),("TAT","ZY")) , defaults to None
        :type route: Tuple[Tuple[str,str]], optional
        :param N: number of qubits, defaults to None
        :type N: int, optional
        :param theta: list of theta angle, defaults to None
        :type theta: List[float], optional
        """        
        self._maxiter = maxiter
        self._t = 0
        self.step_size = lr
        self._eps = eps
        self._tol = tol 
        self.use_qng = use_qng
        if self.use_qng:
            self.route = route
            self.N = N 
            self.theta = theta

    def calc_fubini_tensor(self,params):
        """
        We calculate fubini tensor (note currently only apply for OAT,TNT,TAT gates)

        :param params: list of parameters
        :type params: List
        :return: G - fubini tensor
        :rtype: ndarray, torch, sparse
        """        
        num_layers = len(self.route)-1 
        feature_map = self.route[0][0]
        if torch.is_tensor(params):
            qc = circuit(self.N,use_gpu=True,device=params.device)
            if feature_map == "RN2":
                qc.RN(np.pi/2,0) 
            gs = []
            for num in range(num_layers):
                type = self.route[num+1][1]
                if self.route[num+1][0] == "OAT":
                    g = torch.real(qc.expval("OAT'"+type+"2") - qc.expval("OAT'"+type)**2)
                    qc.OAT(params[num],type)
                elif self.route[num+1][0] == "TAT":
                    g = torch.real(qc.expval("TAT'"+type+"2") - qc.expval("TAT'"+type)**2)
                    qc.TAT(params[num],type)
                elif self.route[num+1][0] == "TNT":
                    g = torch.real(qc.expval("TNT'"+type+"2") - qc.expval("TNT'"+type)**2)
                    qc.TNT(params[num],omega=params[num]*self.N/2,gate_type=type)
                gs.append(g)
            G = torch.block_diag(*gs)
        else:
            qc = circuit(self.N)
            if feature_map == "RN2":
                qc.RN(np.pi/2,0) 
            gs = []
            for num in range(num_layers):
                type = self.route[num+1][1]
                if self.route[num+1][0] == "OAT":
                    g = np.real(qc.expval("OAT'"+type+"2") - qc.expval("OAT'"+type)**2)
                    qc.OAT(params[num],type)
                elif self.route[num+1][0] == "TAT":
                    g = np.real(qc.expval("TAT'"+type+"2") - qc.expval("TAT'"+type)**2)
                    qc.TAT(params[num],type)
                elif self.route[num+1][0] == "TNT":
                    g = np.real(qc.expval("TNT'"+type+"2") - qc.expval("TNT'"+type)**2)
                    qc.TNT(params[num],omega=params[num]*self.N/2,gate_type=type)
                gs.append(g)
            G = block_diag(gs,format="csc")
        return G 

    def optimize(self, num_vars: int, objective_function: Callable[[np.ndarray], float],
                 gradient_function: Optional[Callable[[np.ndarray], float]] = None,
                 initial_point: Optional[np.ndarray] = None,
                 return_loss_hist=None,loss_break=None,return_time_iters=None, 
                 ) -> Tuple[np.ndarray, float, int]:
        """
        :param num_vars: Number of parameters to be optimized.
        :type num_vars: int
        :param objective_function: Handle to a function that computes the objective function.
        :type objective_function: func
        :param gradient_function: Handle to a function that computes the gradient of the objective function., defaults to None
        :type gradient_function: func, optional
        :param initial_point: The initial point for the optimization., defaults to None
        :type initial_point: Optional[np.ndarray], optional
        :param return_loss_hist: return history of loss values, defaults to None
        :type return_loss_hist: bool, optional
        :param loss_break: early stopping, defaults to None
        :type loss_break: bool, optional
        :param return_time_iters: return time each iterations, defaults to None
        :type return_time_iters: bool, optional
        :return: output
        :rtype: Tuple[np.ndarray, float, int]
        """        
        if torch.is_tensor(initial_point):
            pass
        else:
            if initial_point is None:
                initial_point = np.random.default_rng(52).random(num_vars)
            if gradient_function is None:
                gradient_function = lambda params: gradient_num_diff(params,objective_function, self._eps)

        output = self.minimize(objective_function, initial_point, gradient_function,return_loss_hist,return_time_iters,loss_break)
        return output
        
    
    def minimize(self, objective_function: Callable[[np.ndarray], float], initial_point: np.ndarray,
                 gradient_function: Callable[[np.ndarray], float],return_loss_hist:bool,return_time_iters:bool, loss_break:bool) -> Tuple[np.ndarray, float, int]:
        """Run the minimization.

        :param objective_function: A function handle to the objective function.
        :type objective_function: func
        :param initial_point: The initial iteration point.
        :type initial_point: np.ndarray
        :param gradient_function: A function handle to the gradient of the objective function.
        :type gradient_function: func
        :param return_loss_hist: return loss history
        :type return_loss_hist: bool
        :param return_time_iters: return time each iterations
        :type return_time_iters: bool
        :param loss_break: early stopping
        :type loss_break: bool
        :return: A tuple of (optimal parameters, optimal value, number of iterations).
        :rtype: Tuple[np.ndarray, float, int]
        """        
        loss_history = []
        if torch.is_tensor(initial_point):
            tensor_times = []
            objective_function(initial_point).backward()
            derivative = initial_point.grad
            params = params_new = initial_point
            while self._t < self._maxiter:
                start = time.time()
                if self._t > 0:
                    params = params.detach().requires_grad_()
                    objective_function(params).backward()
                    derivative = params.grad
                self._t += 1
                if self.use_qng:
                    G = self.calc_fubini_tensor(params)
                    pinv_G = torch.linalg.pinv(G)
                    params_new = params - self.step_size * pinv_G @ derivative.type(torch.DoubleTensor).to(pinv_G.device)
                else: 
                    params_new = params - self.step_size * derivative
                new_loss = objective_function(params_new)
                loss_history.append(new_loss.item())               
                # print("derivative,params_new,new_loss:",derivative,params_new,new_loss)
                end = time.time()-start
                tensor_times.append(end)
                if torch.linalg.norm(params - params_new) < self._tol:
                    if return_loss_hist or return_time_iters:
                            return params_new, new_loss, self._t, loss_history,tensor_times
                    return params_new,new_loss, self._t 
                elif loss_break != None:
                    if new_loss < loss_break:
                        if return_loss_hist or return_time_iters:
                            return params_new, new_loss, self._t, loss_history, sparse_times
                        return params_new,new_loss, self._t
                    else:
                        params = params_new
                else:
                    params = params_new
            if return_loss_hist or tensor_times:
                        return params_new,new_loss, self._t, loss_history,return_time_iters
            return params_new,new_loss, self._t

        else:
            derivative = gradient_function(initial_point)
            params = params_new = initial_point
            sparse_times = []
            while self._t < self._maxiter:
                start = time.time()
                if self._t > 0:
                    derivative = gradient_function(params)
                self._t += 1
                if self.use_qng:
                    G = self.calc_fubini_tensor(params)
                    pinv_G = linalg.pinv(G.toarray())
                    params_new = params - self.step_size * pinv_G.dot(derivative) 
                else: 
                    params_new = params - self.step_size * derivative
                new_loss = objective_function(params_new)
                loss_history.append(new_loss)
                # print("derivative,params,loss:",derivative,params_new,new_loss)
                end = time.time() - start 
                sparse_times.append(end)
                if np.linalg.norm(params - params_new) < self._tol:
                    if return_loss_hist or return_time_iters:
                            return params_new,new_loss, self._t, loss_history,sparse_times
                    return params_new,new_loss, self._t
                elif loss_break != None:
                    if new_loss < loss_break:
                        if return_loss_hist or return_time_iters:
                            return params_new,new_loss, self._t, loss_history, sparse_times
                        return params_new,new_loss, self._t
                    else:
                        params = params_new
                else:
                    params = params_new
            
            if return_loss_hist or return_time_iters:
                    return params_new,new_loss, self._t, loss_history , sparse_times 
            return params_new,new_loss, self._t

class ADAM:
    """Adam and AMSGRAD optimizers.
    (ref. qiskit)
    """    

    _OPTIONS = ['maxiter', 'tol', 'lr', 'beta_1', 'beta_2',
                'noise_factor', 'eps', 'amsgrad', 'snapshot_dir']

    def __init__(self,
                 maxiter: int = 10000,
                 tol: float = 1e-6,
                 lr: float = 1e-3,
                 beta_1: float = 0.9,
                 beta_2: float = 0.99,
                 noise_factor: float = 1e-8,
                 eps: float = 1e-10,
                 amsgrad: bool = False,
                 snapshot_dir: Optional[str] = None) -> None:
        """

        :param maxiter: Maximum number of iterations, defaults to 10000
        :type maxiter: int, optional
        :param tol: Tolerance for termination, defaults to 1e-6
        :type tol: float, optional
        :param lr: Learning rate, defaults to 1e-3
        :type lr: float, optional
        :param beta_1: Value in range 0 to 1, Generally close to 1., defaults to 0.9
        :type beta_1: float, optional
        :param beta_2: Value in range 0 to 1, Generally close to 1., defaults to 0.99
        :type beta_2: float, optional
        :param noise_factor: Value >= 0, Noise factor, defaults to 1e-8
        :type noise_factor: float, optional
        :param eps: Value >=0, Epsilon to be used for finite differences if no analytic
                gradient method is given., defaults to 1e-10
        :type eps: float, optional
        :param amsgrad: True to use AMSGRAD, defaults to False
        :type amsgrad: bool, optional
        :param snapshot_dir: If not None save the optimizer's parameter after every step to the given directory, defaults to None
        :type snapshot_dir: Optional[str], optional
        """        
        super().__init__()
        self._maxiter = maxiter
        self._snapshot_dir = snapshot_dir
        self._tol = tol
        self._lr = lr
        self._beta_1 = beta_1
        self._beta_2 = beta_2
        self._noise_factor = noise_factor
        self._eps = eps
        self._amsgrad = amsgrad

        # runtime variables
        self._t = 0  # time steps
        self._m = np.zeros(1)
        self._v = np.zeros(1)
        if self._amsgrad:
            self._v_eff = np.zeros(1)

        if self._snapshot_dir:

            with open(os.path.join(self._snapshot_dir, 'adam_params.csv'),
                      mode='w', encoding="utf8") as csv_file:
                if self._amsgrad:
                    fieldnames = ['v', 'v_eff', 'm', 't']
                else:
                    fieldnames = ['v', 'm', 't']
                writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                writer.writeheader()

    def save_params(self, snapshot_dir: str) -> None:
        """
        Save the current iteration parameters to a file called ``adam_params.csv``.
        
        Note:
        
            The current parameters are appended to the file, if it exists already.

            The file is not overwritten.

        :param snapshot_dir: The directory to store the file in.
        :type snapshot_dir: str
        """        
        if self._amsgrad:
            with open(os.path.join(snapshot_dir, 'adam_params.csv'),
                      mode='a', encoding="utf8") as csv_file:
                fieldnames = ['v', 'v_eff', 'm', 't']
                writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                writer.writerow({'v': self._v, 'v_eff': self._v_eff,
                                 'm': self._m, 't': self._t})
        else:
            with open(os.path.join(snapshot_dir, 'adam_params.csv'),
                      mode='a', encoding="utf8") as csv_file:
                fieldnames = ['v', 'm', 't']
                writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                writer.writerow({'v': self._v, 'm': self._m, 't': self._t})

    def load_params(self, load_dir: str) -> None:
        """Load iteration parameters for a file called ``adam_params.csv``.

        :param load_dir: The directory containing ``adam_params.csv``
        :type load_dir: str
        """        

        with open(os.path.join(load_dir, 'adam_params.csv'),
                  mode='r', encoding="utf8") as csv_file:
            if self._amsgrad:
                fieldnames = ['v', 'v_eff', 'm', 't']
            else:
                fieldnames = ['v', 'm', 't']
            reader = csv.DictReader(csv_file, fieldnames=fieldnames)
            for line in reader:
                v = line['v']
                if self._amsgrad:
                    v_eff = line['v_eff']
                m = line['m']
                t = line['t']

        v = v[1:-1]
        self._v = np.fromstring(v, dtype=float, sep=' ')
        if self._amsgrad:
            v_eff = v_eff[1:-1]
            self._v_eff = np.fromstring(v_eff, dtype=float, sep=' ')
        m = m[1:-1]
        self._m = np.fromstring(m, dtype=float, sep=' ')
        t = t[1:-1]
        self._t = np.fromstring(t, dtype=int, sep=' ')

    def minimize(self, objective_function: Callable[[np.ndarray], float], initial_point: np.ndarray,
                 gradient_function: Callable[[np.ndarray], float],return_loss_hist:bool,return_time_iters:bool,loss_break:int) -> Tuple[np.ndarray, float, int]:
        """

        :param objective_function: A function handle to the objective function.
        :type objective_function: func
        :param initial_point: The initial iteration point.
        :type initial_point: np.ndarray
        :param gradient_function: A function handle to the gradient of the objective function.
        :type gradient_function: func
        :param return_loss_hist: return history of losses
        :type return_loss_hist: bool
        :param return_time_iters: return time each iterations
        :type return_time_iters: bool
        :param loss_break: early stopping
        :type loss_break: int
        :return: output
        :rtype: Tuple[np.ndarray, float, int]
        """        
        loss_history = []
        if torch.is_tensor(initial_point):
            tensor_times = []
            objective_function(initial_point).backward()
            derivative = initial_point.grad
            self._t = 0
            self._m = torch.zeros(derivative.size()).to(initial_point.device)
            self._v = torch.zeros(derivative.size()).to(initial_point.device)    
            if self._amsgrad:
                self._v_eff = torch.zeros(derivative.size()).to(initial_point.device)
            params = params_new = initial_point
            while self._t < self._maxiter:
                start = time.time()
                if self._t > 0:
                    params = params.detach().requires_grad_()
                    objective_function(params).backward()
                    derivative = params.grad
                self._t += 1
                self._m = self._beta_1 * self._m + (1 - self._beta_1) * derivative
                self._v = self._beta_2 * self._v + (1 - self._beta_2) * derivative * derivative
                lr_eff = self._lr * math.sqrt(1 - self._beta_2 ** self._t) / (1 - self._beta_1 ** self._t)
                if not self._amsgrad:
                    params_new = (params - lr_eff * self._m.flatten()
                                / (torch.sqrt(self._v.flatten()) + self._noise_factor))
                else:
                    self._v_eff = torch.maximum(self._v_eff, self._v)
                    params_new = (params - lr_eff * self._m.flatten()
                                / (torch.sqrt(self._v_eff.flatten()) + self._noise_factor))
                new_loss = objective_function(params_new)
                loss_history.append(new_loss.item())
                # print("derivative,params,new_loss:",derivative,params_new,new_loss)
                end = time.time() - start 
                tensor_times.append(end)
                if self._snapshot_dir:
                    self.save_params(self._snapshot_dir)
                if torch.linalg.norm(params - params_new) < self._tol:
                    if return_loss_hist or return_time_iters:
                        return params_new, new_loss, self._t, loss_history, tensor_times
                    return params_new,new_loss, self._t
                if loss_break != None:
                    if new_loss < loss_break:
                        if return_loss_hist or return_time_iters:
                            return params_new, new_loss, self._t, loss_history, tensor_times
                        return params_new,new_loss, self._t
                    else:
                        params = params_new
                else:
                    params = params_new

            if return_loss_hist or return_time_iters:
                        return params_new, new_loss, self._t, loss_history, tensor_times
            return params_new, new_loss, self._t


        else:
            derivative = gradient_function(initial_point)
            sparse_times = []
            self._t = 0
            self._m = np.zeros(np.shape(derivative))
            self._v = np.zeros(np.shape(derivative))
            if self._amsgrad:
                self._v_eff = np.zeros(np.shape(derivative))

            params = params_new = initial_point
            while self._t < self._maxiter:
                start = time.time()
                if self._t > 0:
                    derivative = gradient_function(params)
                self._t += 1
                self._m = self._beta_1 * self._m + (1 - self._beta_1) * derivative
                self._v = self._beta_2 * self._v + (1 - self._beta_2) * derivative * derivative
                lr_eff = self._lr * np.sqrt(1 - self._beta_2 ** self._t) / (1 - self._beta_1 ** self._t)
                if not self._amsgrad:
                    params_new = (params - lr_eff * self._m.flatten()
                                / (np.sqrt(self._v.flatten()) + self._noise_factor))
                else:
                    self._v_eff = np.maximum(self._v_eff, self._v)
                    params_new = (params - lr_eff * self._m.flatten()
                                / (np.sqrt(self._v_eff.flatten()) + self._noise_factor))
                new_loss = objective_function(params_new)
                loss_history.append(new_loss)
                # print("derivative,params,new_loss:",derivative,params_new,new_loss)
                end = time.time() - start 
                sparse_times.append(end)
                if self._snapshot_dir:
                    self.save_params(self._snapshot_dir)
                if np.linalg.norm(params - params_new) < self._tol:
                    if return_loss_hist or return_time_iters:
                        return params_new,new_loss, self._t, loss_history, sparse_times
                    return params_new,new_loss, self._t
                if loss_break != None:
                    if new_loss < loss_break:
                        if return_loss_hist or return_time_iters:
                            return params_new,new_loss, self._t, loss_history, sparse_times
                        return params_new,new_loss, self._t
                    else:
                        params = params_new
                else:
                    params = params_new

            if return_loss_hist or return_time_iters:
                        return params_new,new_loss, self._t, loss_history, sparse_times
            return params_new,new_loss, self._t

    def optimize(self, num_vars: int, objective_function: Callable[[np.ndarray], float],
                 gradient_function: Optional[Callable[[np.ndarray], float]] = None,
                 initial_point: Optional[np.ndarray] = None, return_loss_hist = False,
                 loss_break = None,return_time_iters=None
                 ) -> Tuple[np.ndarray, float, int]:
        """Perform optimization.

        :param num_vars: Number of parameters to be optimized.
        :type num_vars: int
        :param objective_function: Handle to a function that computes the objective function.
        :type objective_function: func
        :param gradient_function: Handle to a function that computes the gradient of the objective function, defaults to None
        :type gradient_function: func, optional
        :param initial_point: The initial point for the optimization, defaults to None
        :type initial_point: Optional[np.ndarray], optional
        :param return_loss_hist: return history of loss values, defaults to False
        :type return_loss_hist: bool, optional
        :param loss_break: early stopping, defaults to None
        :type loss_break: bool, optional
        :param return_time_iters: return time each iterations, defaults to None
        :type return_time_iters: bool, optional
        :return: output
        :rtype: Tuple[np.ndarray, float, int]
        """        
        if initial_point is None:
            initial_point = np.random.default_rng(52).random(num_vars)
        if torch.is_tensor(initial_point):
            pass
        else:
            if gradient_function is None:
                gradient_function = lambda params: gradient_num_diff(params,objective_function, self._eps)
        output = self.minimize(objective_function, initial_point, gradient_function,return_loss_hist,return_time_iters,loss_break)
        return output


"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

from numpy import sqrt,real,sum
import numpy as np
from numpy.linalg import eigh
from random import random
from tqix.qtool import dotx
from tqix.qobj import *

__all__ = ['gtrace','gfide','ginfide']

def gtrace(a,b):
    """
    To get trace distance between a and b
    See:  Nielsen & Chuang, "Quantum Computation and Quantum Information"
    
    Input
    ---------
    a : density matrix or state vector
    b : the same as a

    Result
    ---------
    fidelity : float
    """
    # check if they are quantum object
    if isqx(a) and isqx(b):
       a = operx(a)
       b = operx(b)
       if a.shape != b.shape:
          raise TypeError('a and b do not have the same dimensions.')

       diff = a - b
       diff = dotx(daggx(diff),diff)
       # get eigenvalues of sparse matrix
       vals, vecs= eigh(diff)
       return float(real(0.5 * np.sum(sqrt(np.abs(vals)))))
    else:
       msg = 'a or b is not a quantum object'
       raise TypeError(msg)

def gfide(a,b):
    """
    To get fidelity of a and b
    See: Nielsen & Chuang, "Quantum Computation and Quantum Information"
    
    Input
    ---------
    a : density matrix or state vector
    b : the same as a

    Result
    ---------
    fidelity : float
    """

    # check if they are quantum onject
    if isqx(a) and isqx(b):
       if typex(a) != 'oper':
          sqrtma = operx(a)
          if typex(b) != 'oper':
             b = operx(b)
       
       else:
          if typex(b) != 'oper':
             #swap a and b
             return gfide(b,a)
          sqrtma = sqrtx(a)

       if sqrtma.shape != b.shape:
          raise TypeError('Density matrices do not have same dimensions.')
    
       eig_vals, eig_vecs = eigh(dotx(sqrtma,b,sqrtma)) 
       return float(real(sum(sqrt(eig_vals[eig_vals > 0]))))
    else:
        msg = ' a or b is not a quantum object'
        raise TypeError(msg)

def ginfide(a,b):
    #to get infidelity
    return 1-gfide(a,b)



# from tqix import *
# import numpy as np
# from tqix.pis import *

# N = 3

# print('test circuit')
# qc = circuit(N)
# state = qc.state
# print(state)
# print(typex(state))

# print('test gate with initial state')
# qc.RZ(np.pi/3)
# print(qc.state)

from tqix import *
from tqix.pis import *
import numpy as np
N = 10 #qubits
theta=0.4
phi=0.3
qc = circuit(N, use_gpu =False)
qc.RX(np.pi/3,0.05)
# prob = qc.measure(num_shots=1000)
#to get state information
# psi = qc.state #sparse matrix
# psi = qc.state.toarray() #full matrix
# print(prob.sum())


"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['qmeas','qevo','qsim']

import time
from numpy import sqrt,exp,pi
import numpy as np
from tqix.utility import randunit
from tqix.qoper import *
from tqix.qtool import dotx, tensorx
from tqix.qstate import *
from tqix.backend import *
from tqix.qobj import *

from tqix.povm.povm import _pauli,_stoke,_mub,_sic

class qmeas:
    """
    quantum measurement
    """
    def __init__(self,state,*args):
        self.state = state
        self.args = args
    
    def probability(self):
        pr, dt = qmeas._pro_time_(self)
        return pr

    def mtime(self):
        pr, dt = qmeas._pro_time_(self)
        return dt 
    
    def _pro_time_(self):
        if self.state is None:
           msg = 'please input somthing'
           raise TypeError(msg)
        else:
           # turn state into oper
           if typex(self.state) != 'oper':
              instate = operx(self.state)
           else:
              instate = self.state
          
           dims = shapex(instate)[0]
           # check args
           if self.args == None:
              return tracex(dotx(instate,self.args))
           elif self.args[0] == 'Pauli':
              pov = _pauli(int(np.log2(dims)))
           elif self.args[0] == 'Stoke':
              pov = _stoke(int(np.log2(dims)))
           elif self.args[0] == 'MUB':
              pov = _mub(dims)
           elif self.args[0] == 'SIC':
              pov = _sic(dims)
           else: #arbitary POVMs
              if len(self.args) == 1 and \
                 isinstance(self.args[0],(list,np.ndarray)):
                 # e.g. args = [m1,m2,...mn]
                 pov = self.args[0]
              else:
                 pov = self.args

           # now calculate probability
           pr = []
           start = time.time() # starting time
           for i in range (len(pov)):
              pr.append(tracex(dotx(instate,pov[i]))) 
           end = time.time() # stoping time
           dtime = (end - start) # time need to calcualte all proba.   
           return pr, dtime
    
    def poststate(self):
        return self.state #not yet

class qevo:
    # to get the evolution of quantum system 

    def __init__(self,state=None,evol=None,post=None,obser=None):
        self.state = state
        self.evol = evol
        self.post = post
        self.obser = obser

    def output(self):
        # to get output state
        if self.state is None: 
           msg = 'please input somthing'
           raise TypeError(msg)
        else:
           dim = self.state.shape[0]
           if self.evol is None:
              self.evol = eyex(dim)
           if typex(self.state) == 'bra':
              msg = 'state should not a bra vector'
              raise TypeError(msg)
           elif typex(self.state) == 'ket':
              if self.post is None:
                 return(dotx(self.evol,self.state))
              else: # with post-selected
                 if typex(self.post) != 'bra':
                    msg = 'post state must be a bra vector'
                    raise TypeError(msg)
                 else:
                    return(dotx(self.post,self.evol,self.state))
           else: # state is an oper
              if self.post is None:
                 return(dotx(self.evol,self.state,daggx(self.evol)))
              else: # with post-selected
                 if typex(self.post) != 'bra':
                    msg = 'post state must be a bra vector'
                    raise TypeError(msg)
                 else:
                    pp = dotx(self.evol,self.state,dagg(self.evol))
                    return(dotx(self.post,pp,dagg(self.post)))

    def info(self):
       print('to get measurement results: outcome state,expectation')

class qsim:
    # to simulate a measuremet
    def __init__(self,qmeas,niter = 1000,backend='cdf'):
        self.qmeas = qmeas.probability()
        self.niter = niter
        self.backend = backend
 
    def get_qmeas_simul(self):
        res = []
        for i in range(len(self.qmeas)):
            if self.backend == 'mc':
               res.append(mc(np.real(self.qmeas[i]),self.niter))
            elif self.backend == 'cdf':
               temp = []
               for j in range(self.niter):
                   temp.append(randunit()/np.real(self.qmeas[i]))
               res.append(cdf(temp))
            else:
               raise TypeError('no ', backend) 
        return res    




# -*- coding: utf-8 -*-
""" Entanglement measure
"""

import qiskit
import tqix
import numpy as np
from math import comb

def concentratable_entanglement(
    qc: qiskit.QuantumCircuit):
    """Calculate Concentratable Entanglement (CE)

    Args:
        qc (qiskit.QuantumCircuit): circuit to be calculated

    Returns:
        (complex): CE
    """
    
    n_qubit = qc.num_qubits #this is ancilar
    swap_qc = create_swap_test_state(qc.copy())
    
    list_measured = list(range(0, n_qubit))
    prob = tqix.vqa.circuits.measure(swap_qc, list_measured)
    return 1 - prob

def concentratable_entanglement_theor(
    qc: qiskit.QuantumCircuit):
    """Calculate Concentratable Entanglement (CE)

    Args:
        qc (qiskit.QuantumCircuit): circuit to be calculated

    Returns:
        (complex): CE
    """
    
    n_qubit = qc.num_qubits #this is ancila qubits
    swap_qc = create_swap_test_state(qc.copy()) #full qubits
    
    list_measured = list(range(0, n_qubit))
    prob = tqix.vqa.circuits.measure_theor(swap_qc, list_measured)
    return 1 - prob

    
def create_swap_test_state(
    qc: qiskit.QuantumCircuit):
    """Create swap test state

    Args:
        qc (qiskit.QuantumCircuit): init circuit

    Returns:
        (qiskit.QuantumCircuit): swap test circuit
    """
    
    n_qubit = qc.num_qubits
    qubits_list_first = list(range(n_qubit, 2*n_qubit))
    qubits_list_second = list(range(2*n_qubit, 3*n_qubit))
    
    # Create swap test circuit
    swap_test_circuit = qiskit.QuantumCircuit(3*n_qubit, n_qubit)
    
    # Add initial circuit the first time
    swap_test_circuit = swap_test_circuit.compose(qc, qubits=qubits_list_first)
    # Add initial circuit the second time
    swap_test_circuit = swap_test_circuit.compose(qc, qubits=qubits_list_second)
    swap_test_circuit.barrier()
    
    # Add hadamard gate
    swap_test_circuit.h(list(range(0, n_qubit)))
    swap_test_circuit.barrier()
        
    for i in range(n_qubit):
        # Add control-swap gate
        swap_test_circuit.cswap(i, i+n_qubit, i+2*n_qubit)
    swap_test_circuit.barrier()
    
    # Add hadamard gate
    swap_test_circuit.h(list(range(0, n_qubit)))
    swap_test_circuit.barrier()
    
    return swap_test_circuit

def ce_max(n):
    """
    return maximun CE with input n-qubit
    Ref: arXiv:2207.11997v1 (2022)
    """
    ce = 0
    for j in range (0,n+1):
        ce += comb(n,j)/2**np.min([j,n-j])
    ce = ce/2**n
    ce = 1-ce
    return ce



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: Quangtuan Kieu
>>> all rights reserved
________________________________
"""

__all__ = ['bx','bz',
            'obasis','dbasis','zbasis','dzbasis',
            'coherent','squeezed', 'position','spin_coherent',
            'ghz', 'w', 'dicke', 'random', 
            'add_random_noise','add_white_noise',
            'ghz_minmax', 'star']

from numpy import conj, kron, sqrt, exp, pi, sin, cos

import numpy as np
from itertools import combinations
from scipy.special import comb,factorial
from tqix.utility import randunit,randnormal,haar
from tqix.qtool import dotx
from tqix.qoper import *
from tqix.qobj import *

def bx(d,e = 0):
    """ to generate an orthogonal basis of d dimension at e
    For example:
    base(2,0) = [[1]
                [0]
                [0]]
    """
    if (not isinstance(d,int)) or d < 0:
        raise ValueError("d must be integer d >= 0")
    if (not isinstance(e,int)) or e < 0:
        raise ValueError("e must be interger e>= 0")
    if e > d-1:
        raise ValueError("basis vector index need to be in d-1")
    ba = np.zeros((d,1)) #column vector
    ba[e,0] = 1.0
    return qx(ba)

def obasis(d,e = 0):
    #old version
    if (not isinstance(d,int)) or d < 0:
        raise ValueError("d must be integer d >= 0")
    if (not isinstance(e,int)) or e < 0:
        raise ValueError("e must be interger e>= 0")
    if e > d-1:
        raise ValueError("basis vector index need to be in d-1")
    #print('Warnings: obasis(d,e) is an old version, please use bx(d,e) instead.')
    ba = np.zeros((d,1)) #column vector
    ba[e,0] = 1.0
    return qx(ba)

def dbasis(d,e = 0):
    # generate a dual basis
    return daggx(bx(d,e))

def bz(j,m):
    """to generate Zeeman basis |j,m> of Sz spin operator
    ie., Sz|j,m> = j|j,m>
    Parameters
    -----------
    j: real number
    m: integer number
    """
    return qx(bx(int(2*j+1),int(j-m)))

def zbasis(j,m):
    #old version
    #print('Warnings: zbasis(j,m) is an old version, please use bz(d,e) instead.')
    return qx(bx(int(2*j+1),int(j-m)))

def dzbasis(j,m):
    # generate a dual Zeeman basis
    return daggx(bz(j,m))

def coherent(d,alpha):
    """ to generate coherent state
    Parameters
    ----------
    alpha: complex number
    d: number of dimensions
    """
    ground = bx(d,0)
    coherent = dotx(displacement(d,alpha), ground)
    coherent = normx(coherent)
    return qx(coherent)
    
def squeezed(d,alpha,beta):
    """ to generate squeezed state
    Parameters
    ----------
    alpha: complex number
    d: number of dimensions
    """
    ground = bx(d,0)
    squeezed = dotx(displacement(d,alpha),
                squeezing(d,beta),ground)
    squeezed = normx(squeezed)
    return qx(squeezed)

def position(d,x):
    """ to generate position state
    ie., X|x> = x|x>
    Parameters
    -----------
    x: position coordinate, real number
    d: number of dimensions
    """
    ground = bx(d,0)
    power = -0.5 * dotx(raising(d),raising(d)) + \
            sqrt(2) * x * raising(d)
    exponential = expx(power)
    position = exp(-x**2/2) /(pi ** (1./4)) * dotx(exponential, ground)
    position = normx(position)
    return qx(position)

def spin_coherent(j,theta,phi):
    """ to generate spin coherent state 
    defined as |j,theta,phi> = exp(i*phi*Sz)*exp(i*theta*Sy)|j,j>
    Parameters
    -----------
    j: real number
    theta: polar angle
    phi: azimuthal angle
    z_rotate = expx(1j * phi * soper(j,'z'))
    y_rotate = expx(1j * theta * soper(j,'y'))
    total_rotate = dotx(z_rotate,y_rotate)
    return qx(dotx(total_rotate,spinx(j,j)))
    """
    state = 0.0
    array = np.arange(-j,j+1,1)
    for m in array:
        state += sqrt(factorial(2*j)/(factorial(j+m)*factorial(j-m)))\
                * (cos(theta/2.0))**(j+m) * (sin(theta/2.0))**(j-m)\
                * exp(-1j*m*phi) * bz(j,m)
    return state

def ghz(n):
    """ to generate GHZ state
    Parameters
    ----------
    n: number of qubits

    Return: GHZ state, 
    ie. (|00...0> + |11...1>)/sqrt(2)

    """
    dim = 2**n
    up,down = _up_down()
    ups,upd = up,down
    for i in range(n-1):
        ups = kron(ups,up)
        upd = kron(upd,down)
    GHZ = (ups+upd)/sqrt(2.0)
    return qx(GHZ)

def w(n):
    """ to generate W state
    Parameters
    ----------
    n: number of qubits

    Return: dicke(n,1)

    """
    return dicke(n,1)

def dicke(n,e):
    """ to generate Dicke state
    Parameters
    ----------
    n: number of qubits
    e: excited qubits

    Return: dicke state

    """
    dim = 2**n
    comb_array = _place_ones(n,e)
    #array of possible values of qubits
    row, col = comb_array.shape
    temp_sum = np.zeros((dim,1))

    for i in range(row):
        temp_vec = _upside_down(comb_array[i,0])
        for j in range(1,col):
            temp_vec0 = kron(temp_vec,_upside_down(comb_array[i,j]))
            temp_vec = temp_vec0
        temp_sum += temp_vec

    dicke = temp_sum/(np.sqrt(row))
    return qx(dicke)

def random(d):
    """ to generate a random state
    Parameters:
    ----------
    d: dimension

    Return: random state

    """
    rpsi = np.zeros((d,1))
    ipsi = np.zeros((d,1))
    for i in range (d):
        rpsi[i,0] = randunit()
        ipsi[i,0] = randunit()
    ppsi = rpsi + 1j*ipsi
    ppsi = normx(ppsi)
    M = haar(d)
    prime = dotx((eyex(d)+M),ppsi)
    prime = normx(prime)
    return qx(prime)    

def add_random_noise(psi,m = 0.0,st = 0.0):
    """ to generate 1 perturbed random state from psi
    Parameters:
    ----------
    d: dims
    m: mean
    st: standard derivative
    """
    if isqx(psi):
        dim = psi.shape[0]
        per = [randnormal(m,st,dim)+1j*randnormal(m,st,dim)]
        if typex(psi)=='ket':
            per = daggx(per) #to get ket
        elif typex(psi)=='oper':
            per = dotx(daggx(per),per)

        psi = psi + per
        psi = normx(psi)
        return qx(psi)
    else:
        msg = 'psi is not a quantum object'
        raise TypeError(msg)

def add_white_noise(state,p = 0.0):
    """ add white noise to quantum state
    Parameters:
    ----------
    state: quantum state
    p: error
    
    Return
    ------
    (1-p)*state + p*I/d
    """
    if typex(state) != 'oper':
        state = operx(state)
    dim = state.shape[0]
    return qx((1-p)*state+p*eyex(dim)/dim)

def ghz_minmax(A):
    """to create a ghz-like state with 
    min-max eigenvalues of operator A

    Args:
        A (complex): operator

    Raises:
        TypeError: if A is not an oper

    Returns:
        matrix: ghz state
    """
    # check is A is an oper or not
    if typex(A) != 'oper':
        msg = 'A must be an operator'
        raise TypeError(msg)
    
    # get min-max eigenvalues
    w,vk = eigenx(A)
    idx_min = np.argmin(w)
    idx_max = np.argmax(w)
    return (1/np.sqrt(2.0)*(vk[idx_min]+vk[idx_max]))

def star(n):
    """ to generate star state
    Parameters
    ----------
    n: number of qubits

    Return: GHZ state, 
    ie. (|0+...+> + |1-...->)/sqrt(2)

    """
    dim = 2**n
    up,down = _up_down()
    plus = normx(up + down)
    minus = normx(up - down)
    
    ups,upd = up,down
    for i in range(n-1):
        ups = kron(ups,plus)
        upd = kron(upd,minus)
    st = (ups+upd)/sqrt(2.0)
    return qx(st)

#####
def _up_down():
    # to generate up and down state
    up = np.zeros((2,1))
    up[0,0] = 1.0
    down = np.zeros((2,1))
    down[1,0] = 1.0
    return qx(up),qx(down)

def _upside_down(a):
    """
    return |0> or |1> or none
    """
    up, down = _up_down()
    if a == 0:
        return up
    elif a == 1:
        return down
    else:
        msg = 'Out of acceptable range'
        raise TypeError(msg)

#from math import comb
def _place_ones(size,count):
    """
    return an array of possible values of Dicke state
    Eg.
    place_ones(3,2) <=> |110>, |101> |011>
    [[1. 1. 0.]
    [1. 0. 1.]
    [0. 1. 1.]]
    """
    c = int(comb(size,count))
    result = np.zeros((c,size))
    k = 0
    for positions in combinations(range(size),count):
        for i in positions:
            result[k,i] = 1
        k += 1
    return result




"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

import os
import sys
#import warnings

#import tqix.version
from tqix.version import version as __version__
import sys

# ------------------------------------------
# Load modules
# ==========================================
# utility

from tqix.utility import *
from tqix.about import *

# outer loop
from tqix.qstate import *
from tqix.qobj import *
from tqix.qmeas import *
from tqix.qoper import *
from tqix.qtool import *
from tqix.backend import *
from tqix.qmetro import *

from tqix.quasi_prob import *
from tqix.visualize import *

# quantum state tomography
from tqix.dsm import *

# povms
from tqix.povm import *

# pis
from tqix.pis import *

# vqa
from tqix.vqa import *



# -----------------------------------------------------------------------------
# Clean name space
#
del os, sys#, numpy, scipy, multiprocessing, distutils



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""

__all__ = ['randunit','krondel','randnormal','haar','ndiff']

import numpy as np
import random
from tqix.qobj import *
from scipy import (randn,diagonal,absolute,multiply)
from scipy.linalg import qr

def randunit():
    # to generate a random number [0,1]
    return random.random()

def krondel(i,j):
    # to generate kronecker delta
    return 1 if i==j else 0

def randnormal(m=0.0,s=0.0,n=1):
    """
    to generate normal distribution
    
    Parameters:
    -----------
    m: mean
    s: standard deviation
    n: size
    """
    return np.random.normal(m,s,n)
    
def haar(d):
    """
    to generate random matrix for Haar measure
    see https://arxiv.org/pdf/math-ph/0609050.pdf
    """
    array = (randn(d,d) + 1j*randn(d,d))/np.sqrt(2.0)
    ortho,upper = qr(array)
    diag = diagonal(upper)
    temp = diag/absolute(diag)
    result = multiply(ortho,temp,ortho)
    return result

def ndiff(x,y):
    """
    to calculate numerical differential of y(x)
    input: x, y: arrays,
    output: dy: differential
    """
    dy = np.zeros(x.shape,float)
    dy[0:-1] = np.diff(y)/np.diff(x)
    dy[-1] = (y[-1]-y[-2])/(x[-1]-x[-2])
    return dy



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                                x: quantum measurement, quantum metrology, 
                                quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> all rights reserved
________________________________
"""
# for quantum metrology calculated from theory

__all__ = ['qfimx', 'qboundx', 'Hx', 'Ux', 'Ax']

import numpy as np
from tqix.qobj import daggx, dotx, eigenx, operx, tracex
from tqix.qoper import eyex
from scipy.linalg import expm,inv
from scipy import sparse

def qfimx(inp_state,h_opt,c_opt,t):
    """calculate quantum fisher information

    Args:
        inp_state (matrix): input quantum state 
        h_opt(list): list of Hamiltonian
        c_opt(list): list of coefficients
        t(float): time
        
    Returns:
        qfim: quantum fisher information matrix    
    """
    
    # length of parameters
    d = len(h_opt)
    
    # get rho
    if isinstance(inp_state, sparse.spmatrix):
        inp_state = inp_state.toarray()
    rho = operx(inp_state)
    
    # unitary, #Ax #final state
    uni = Ux(h_opt,c_opt,t)
    ax = Ax(h_opt,c_opt,t)
    fin_rho = dotx(uni,rho,daggx(uni))
    
    # eigen_rho,  # derivative of fin_state
    eigens = eigenx(fin_rho)
    drhos = _droh(rho,uni,ax) #use inp_rho to calculate drho,
    
    # qfim
    Q = np.zeros((d,d), dtype = complex)  
    for k in range(d):
        for l in range(d):
            Q[k,l] = 0.0
            for i in range(len(eigens[0])):
                for j in range(len(eigens[0])):
                    de = eigens[0][i] + eigens[0][j]
                    if de > 10e-15:
                        num1 = dotx(daggx(eigens[1][i]),drhos[k],eigens[1][j])
                        num2 = dotx(daggx(eigens[1][j]),drhos[l],eigens[1][i])
                        Q[k,l] += 2*np.real(num1*num2)/de

    return np.real(Q)     

def qboundx(inp_state,h_opt,c_opt,t):
    """calculate quantum bound

    Args:
        inp_state (matrix): input quantum state 
        h_opt(list): list of Hamiltonian
        c_opt(list): list of coefficients
        t(float): time
        
    Returns:
        qbound: quantum bound   
    """
    d = len(h_opt)
    W = eyex(d)
    qfim = qfimx(inp_state,h_opt,c_opt,t)
    result = np.real(tracex(W @ inv(qfim + W * 10e-10)))
    
    return result


def Hx(h_opt,c_opt):
    """to create Hamiltonian from h_opt and c_opt

    Args:
        h_opt (list): list of hamiltonian 
        c_opt (list): list of coefficients

    Returns:
        Hx: Hamiltonian
    """
    if len(h_opt) != len(c_opt):
        raise TypeError("list length does not match")
    else:
        H = 0.0
        for i in range(len(h_opt)):
            H += h_opt[i]*c_opt[i]
        return H    

def Ux(h_opt,c_opt,t):
    """to create Unitary operator from h_opt and c_opt

    Args:
        h_opt (list): list of hamiltonian 
        c_opt (list): list of coefficients
        t (float): time

    Returns:
        Ux: Unitary operator
    """
    return expm(-1j*t*Hx(h_opt,c_opt))
            
###
#def _drho():
    
def Ax(h_opt,c_opt,t):
    """calculate Ax = exp(itH)@A@exp(-itH)

    Args:
        h_opt (list): list of Hamiltonian
        c_opt (list): list of coefficients
        t (float): time

    Returns:
        Ax: list of Ax
    """
    if len(h_opt) != len(c_opt):
        raise TypeError("list length does not match")
    
    axl = []
    hamil = Hx(h_opt,c_opt)
    xv = np.linspace(0,t,1000)
    for i in range(len(h_opt)):
        f = lambda u: dotx(expm(1j*u*hamil),h_opt[i],expm(-1j*u*hamil))
        result = np.apply_along_axis(f,0,xv.reshape(1,-1))
        axl.append(np.trapz(result,xv))
    return axl #list of Ax    

def _droh(rho,U,A):
    """calculate d_rho following Eq.(C1) PRA 102, 022602 (2020)

    Args:
        rho (matrix): quantum state need to calculate deri
        U (matrix): unitary
        A (list): list of Ak
    """
    drhos = []
    for i in range(len(A)):
        dU = -1j*dotx(U,A[i])
        dUU = dotx(dU,rho,daggx(U))
        UdU = dotx(U,rho,daggx(dU))
        drhos.append(dUU + UdU)
    return drhos    



from tqix import *
from tqix.qtool import dotx,tensorx
from itertools import product

__all__ = ['_pauli_']

def _pauli_(n):
    """
    Pauli POVM

    Parameters:
    ----------
    n : number of qubits
    
    Return:
    ------
    set of POVM accroding Pauli matrics
    """
    squbit = _pauli_eigens()
    nqubit = []
    T = []
    T = list(product((squbit),repeat = n))
    for i in range(len(T)):
        M = []
        for j in T[i]:
           M.append(j)
        nqubit.append(tensorx(M))
    return nqubit

def _pauli_eigens():
    """
    to get eigenstate of pauli matrices
    there are 3 bases: {0, 1}, {plus, minus}, {L, R}

    Return:
    -------
    Pauli eigenstate POVM set
    """
    u = obasis(2,0)
    d = obasis(2,1)
    h = dotx(u,daggx(u))
    v = dotx(d,daggx(d))
    p = dotx(normx(u+d),daggx(normx(u+d)))
    m = dotx(normx(u-d),daggx(normx(u-d)))
    l = dotx(normx(u+1j*d),daggx(normx(u+1j*d)))
    r = dotx(normx(u-1j*d),daggx(normx(u-1j*d)))
    m = [h,v,p,m,l,r]
    return m




import numpy as np
from tqix.qobj import *
from tqix.pis.util import *
from tqix.pis import *
from scipy.sparse import csc_matrix,lil_matrix
import multiprocessing 
import functools 
import torch 

__all__ =['add_noise','calc_rho_0']

def calc_rho_0(rho_0,iks,jmm1,state,all_iks,j_min,j_max,N_in,d_dicke,use_gpu=False,result_queue=None):
    """

    :param rho_0: initial rho_0 state
    :type rho_0: ndarray,tensor,sparse
    :param iks: list of i,k indexes 
    :type iks: List[tuple]
    :param jmm1: dictionary of i,k keys and j,m,m1 values
    :type jmm1: dict
    :param state: input state
    :type state: ndarray, tensor, sparse
    :param all_iks: dictionary of j,m,m1 keys and i,k values
    :type all_iks: dict
    :param j_min: minimum j index
    :type j_min: int
    :param j_max: maximum j index
    :type j_max: int
    :param N_in: number of qubits 
    :type N_in: int
    :param d_dicke: dimension of dicke basis 
    :type d_dicke: int
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param result_queue: List of queues processed, defaults to None
    :type result_queue: List, optional
    :return: calculated rho_0 state
    :rtype: ndarray, tensor, sparse 
    """        
    if use_gpu:
        accumulate_states = rho_0
        for ik in iks:
            i,k = ik
            j,m,m1 = jmm1[i,k]
            first_term = 0
            second_term = 0
            third_term = 0
            gamma_1 = 0; gamma_2 = 0; gamma_3 = 0; gamma_4 = 0; gamma_5 = 0; gamma_6 = 0
            gamma_7 = 0; gamma_8 = 0; gamma_9 = 0
            p_jmm1 = state[i,k]

            if j >= j_min and j <= j_max and m > -j and m1 > -j and m < j and m1 < j:
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m,m1]] = 1 
                gamma_1 = m*m1*Lambda_a*dicke_bx_m

            if (j-1) >= j_min and (j-1) <= j_max and (m >= -(j-1) and m1 >= -(j-1)) and (m <= (j-1) and m1 <= (j-1)):
                B_jm_z = get_B(j,m,type="z")
                B_jm1_z = get_B(j,m1,type="z")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m,m1]] = 1 
                gamma_2 = B_jm_z*B_jm1_z*Lambda_b*dicke_bx_m

            if (j+1) <= j_max and (j+1) >= j_min and (m >= -(j+1) and m1 >= -(j+1)) and (m <= (j+1) and m1 <= (j+1)):
                D_jm_z = get_D(j,m,type="z")
                D_jm1_z = get_D(j,m1,type="z")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m,m1]] = 1
                gamma_3 = D_jm_z*D_jm1_z*Lambda_d*dicke_bx_m
            
            first_term = (gamma_1+gamma_2+gamma_3)
            
            if j >= j_min and j <= j_max  and (m-1) >= -j and (m1-1) >= -j and (m-1) <= j and (m1-1) <= j:
                A_jm_minus = get_A(j,m,type="-")
                A_jm1_minus = get_A(j,m1,type="-")
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m-1,m1-1]] = 1
                gamma_4 = A_jm_minus*A_jm1_minus*Lambda_a*dicke_bx_m
            
            if (j-1) >= j_min and (j-1) <= j_max and (m-1) >= -(j-1) and (m1-1) >= -(j-1) and (m-1) <= (j-1) and  (m1-1) <= (j-1):
                B_jm_minus = get_B(j,m,type="-")
                B_jm1_minus = get_B(j,m1,type="-")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m-1,m1-1]] = 1
                gamma_5 = B_jm_minus*B_jm1_minus*Lambda_b*dicke_bx_m
            
            if  (j+1) >= j_min and (j+1) <= j_max and (m-1) >= -(j+1) and (m1-1) >= -(j+1) and (m-1) <= (j+1) and (m1-1)  <= (j+1):
                D_jm_minus = get_D(j,m,type="-")
                D_jm1_minus = get_D(j,m1,type="-")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m-1,m1-1]] = 1
                gamma_6 = D_jm_minus*D_jm1_minus*Lambda_d*dicke_bx_m
            second_term = (gamma_4+gamma_5+gamma_6)/2
            
            if j >= j_min and j <= j_max and (m+1) <= j and (m1+1) <= j and (m+1) >= -j and (m1+1) >= -j:
                A_jm_plus = get_A(j,m,type="+")
                A_jm1_plus = get_A(j,m1,type="+")
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m+1,m1+1]] = 1
                gamma_7 = A_jm_plus*A_jm1_plus*Lambda_a*dicke_bx_m
            
            if (j-1) >= j_min and (j-1) <= j_max and (m+1) >= -(j-1) and (m1+1) >= -(j-1) and (m+1) <= (j-1) and (m1+1) <= (j-1):
                B_jm_plus = get_B(j,m,type="+")
                B_jm1_plus = get_B(j,m1,type="+")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m+1,m1+1]] = 1
                gamma_8 = B_jm_plus*B_jm1_plus*Lambda_b*dicke_bx_m
            
            if (j+1) >= j_min and (j+1) <= j_max and  (m+1) >= -(j+1) and (m1+1) >= -(j+1) and (m+1) <= (j+1) and (m1+1) <= (j+1):
                D_jm_plus = get_D(j,m,type="+")
                D_jm1_plus = get_D(j,m1,type="+")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m+1,m1+1]] = 1
                gamma_9 = D_jm_plus*D_jm1_plus*Lambda_d*dicke_bx_m

        third_term = (gamma_7+gamma_8+gamma_9)/2
        accumulate_states += p_jmm1*(first_term+second_term+third_term)
        if result_queue is not None:
            result_queue.put(accumulate_states)
            result_queue.join()
        else:
            return accumulate_states
    else:
        accumulate_states = csc_matrix((d_dicke, d_dicke), dtype=np.complex)
        for ik in iks:
            i,k = ik 
            j,m,m1 = jmm1[i,k]
            first_term = 0
            second_term = 0
            third_term = 0
            gamma_1 = 0; gamma_2 = 0; gamma_3 = 0; gamma_4 = 0; gamma_5 = 0; gamma_6 = 0
            gamma_7 = 0; gamma_8 = 0; gamma_9 = 0
            p_jmm1 = state[i,k]

            if j >= j_min and j <= j_max and m > -j and m1 > -j and m < j and m1 < j:
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m,m1]] = 1 
                gamma_1 = m*m1*Lambda_a*dicke_bx_m

            if (j-1) >= j_min and (j-1) <= j_max and (m >= -(j-1) and m1 >= -(j-1)) and (m <= (j-1) and m1 <= (j-1)):
                B_jm_z = get_B(j,m,type="z")
                B_jm1_z = get_B(j,m1,type="z")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m,m1]] = 1
                gamma_2 = B_jm_z*B_jm1_z*Lambda_b*dicke_bx_m

            if (j+1) <= j_max and (j+1) >= j_min and (m >= -(j+1) and m1 >= -(j+1)) and (m <= (j+1) and m1 <= (j+1)):
                D_jm_z = get_D(j,m,type="z")
                D_jm1_z = get_D(j,m1,type="z")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m,m1]] = 1               
                gamma_3 = D_jm_z*D_jm1_z*Lambda_d*dicke_bx_m
            
            first_term = (gamma_1+gamma_2+gamma_3)
            
            if j >= j_min and j <= j_max  and (m-1) >= -j and (m1-1) >= -j and (m-1) <= j and (m1-1) <= j:
                A_jm_minus = get_A(j,m,type="-")
                A_jm1_minus = get_A(j,m1,type="-")
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m-1,m1-1]] = 1 
                gamma_4 = A_jm_minus*A_jm1_minus*Lambda_a*dicke_bx_m
            
            if (j-1) >= j_min and (j-1) <= j_max and (m-1) >= -(j-1) and (m1-1) >= -(j-1) and (m-1) <= (j-1) and  (m1-1) <= (j-1):
                B_jm_minus = get_B(j,m,type="-")
                B_jm1_minus = get_B(j,m1,type="-")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m-1,m1-1]] = 1          
                gamma_5 = B_jm_minus*B_jm1_minus*Lambda_b*dicke_bx_m
            
            if  (j+1) >= j_min and (j+1) <= j_max and (m-1) >= -(j+1) and (m1-1) >= -(j+1) and (m-1) <= (j+1) and (m1-1)  <= (j+1):
                D_jm_minus = get_D(j,m,type="-")
                D_jm1_minus = get_D(j,m1,type="-")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m-1,m1-1]] = 1             
                gamma_6 = D_jm_minus*D_jm1_minus*Lambda_d*dicke_bx_m
            
            second_term = (gamma_4+gamma_5+gamma_6)/2
            
            if j >= j_min and j <= j_max and (m+1) <= j and (m1+1) <= j and (m+1) >= -j and (m1+1) >= -j:
                A_jm_plus = get_A(j,m,type="+")
                A_jm1_plus = get_A(j,m1,type="+")
                Lambda_a = get_Lambda(N_in,j,type="a")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j,m+1,m1+1]] = 1  
                gamma_7 = A_jm_plus*A_jm1_plus*Lambda_a*dicke_bx_m
            
            if (j-1) >= j_min and (j-1) <= j_max and (m+1) >= -(j-1) and (m1+1) >= -(j-1) and (m+1) <= (j-1) and (m1+1) <= (j-1):
                B_jm_plus = get_B(j,m,type="+")
                B_jm1_plus = get_B(j,m1,type="+")
                Lambda_b = get_Lambda(N_in,j,type="b")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j-1,m+1,m1+1]] = 1
                gamma_8 = B_jm_plus*B_jm1_plus*Lambda_b*dicke_bx_m
            
            if (j+1) >= j_min and (j+1) <= j_max and  (m+1) >= -(j+1) and (m1+1) >= -(j+1) and (m+1) <= (j+1) and (m1+1) <= (j+1):
                D_jm_plus = get_D(j,m,type="+")
                D_jm1_plus = get_D(j,m1,type="+")
                Lambda_d = get_Lambda(N_in,j,type="d")
                dicke_bx_m = rho_0
                dicke_bx_m[all_iks[j+1,m+1,m1+1]] = 1
                gamma_9 = D_jm_plus*D_jm1_plus*Lambda_d*dicke_bx_m

            third_term = (gamma_7+gamma_8+gamma_9)/2
            accumulate_states += p_jmm1*(first_term+second_term+third_term)

        return accumulate_states

def add_noise(qc,noise=0.3,num_process=None,use_gpu=False,device=None):
    """

    :param qc: circuit object
    :type qc: Circuit
    :param noise: noise ratio to add, defaults to 0.3
    :type noise: float, optional
    :param num_process: number of process for multi-processing, defaults to None
    :type num_process: int, optional
    :param use_gpu: if use gpu, defaults to False
    :type use_gpu: bool, optional
    :param device: name of compute device, defaults to None
    :type device: str, optional
    :return: new state after being added noise
    :rtype: ndarray, tensor, sparse
    """    

    state = qc.state
    d_in = shapex(state)[0]
    N_in = qc.N
    d_dicke = get_dim(N_in)

    if d_in != d_dicke:
        if use_gpu:
            state = torch.nn.functional.pad(state,(0,d_dicke-d_in,0,d_dicke-d_in)) 
        else:
            state = lil_matrix(np.pad(state.toarray(),((0,d_dicke-d_in),(0,d_dicke-d_in))))
    new_Nds = get_Nds(shapex(state)[0])

    assert N_in == new_Nds, "not full block"
    if use_gpu:
        non_zero_arrays = state.nonzero(as_tuple=True)
        iks = list(zip(non_zero_arrays[0].detach().cpu().tolist(),non_zero_arrays[1].detach().cpu().tolist()))
    else:
        non_zero_arrays = state.nonzero()   
        iks = list(zip(non_zero_arrays[0],non_zero_arrays[1]))
    
    jmm1,all_iks  = get_jmm1_idx(N_in)
    if use_gpu:
        rho_0 = torch.zeros(d_dicke, d_dicke).type(torch.complex128).to(device)
        if num_process is not None:
            rho = state.detach().cpu()
        else:
            rho = state
    else:
        rho_0 = csc_matrix((d_dicke, d_dicke), dtype=np.complex)
        rho = state
    j_min = get_jmin(N_in)
    j_max = N_in/2

    run_arguments = []
    len_iks = len(iks)
    if num_process is None:
        pass
    else:
        for i in range(num_process):    
            begin_idx = round(len_iks*i/num_process)
            end_idx = round(len_iks*(i+1)/num_process)
            if use_gpu:
                run_arguments.append((rho_0.detach().cpu(),iks[begin_idx:end_idx],jmm1,state.detach().cpu(),all_iks,j_min,j_max,N_in,d_dicke,use_gpu))
            else:
                run_arguments.append((rho_0,iks[begin_idx:end_idx],jmm1,state,all_iks,j_min,j_max,N_in,d_dicke))

    if use_gpu:
        if num_process is None:
            rho_0 = calc_rho_0(rho_0,iks,jmm1,state,all_iks,j_min,j_max,N_in,d_dicke,use_gpu)
            rho_0 = 4/(3*N_in)*rho_0
            normalized_rho_0 = daggx(rho_0) @ (rho_0)/((daggx(rho_0) @ (rho_0)).diagonal().sum())
            new_state = (1-noise)*rho + noise*normalized_rho_0 
            return new_state
        else:
            import torch.multiprocessing as mp
            processes = []
            accumulate_states=[]
            result_queue = mp.JoinableQueue()
            for rank in range(num_process):
                arg_list = list(run_arguments[rank])
                arg_list.append(result_queue)
                run_arguments[rank] = tuple(arg_list)
                p = mp.Process(target=calc_rho_0, args=run_arguments[rank])
                p.start()
                temp_result = result_queue.get()
                result_queue.task_done()
                accumulate_states.append(temp_result)
                del temp_result
                processes.append(p)
            for p in processes:
                p.join()
            rho_0 = functools.reduce(lambda x,y: x+y,accumulate_states)
            rho_0 = 4/(3*N_in)*rho_0
            normalized_rho_0 = daggx(rho_0) @ (rho_0)/((daggx(rho_0) @ (rho_0)).diagonal().sum())
            new_state = (1-noise)*rho + noise*normalized_rho_0 
            return new_state.to(device)
    else:
        if num_process is None:
            rho_0 = calc_rho_0(rho_0,iks,jmm1,state,all_iks,j_min,j_max,N_in,d_dicke,use_gpu)
            rho_0 = 4/(3*N_in)*rho_0
            normalized_rho_0 = daggx(rho_0).dot(rho_0)/((daggx(rho_0).dot(rho_0)).diagonal().sum())
            new_state = (1-noise)*rho + noise*normalized_rho_0             
            return new_state

        else:
            pool = multiprocessing.Pool(processes=num_process)
            accumulate_states = pool.starmap(calc_rho_0,run_arguments)
            pool.close()
            pool.join()    
            rho_0 = functools.reduce(lambda x,y: x+y,accumulate_states)
            rho_0 = 4/(3*N_in)*rho_0
            normalized_rho_0 = daggx(rho_0).dot(rho_0)/((daggx(rho_0).dot(rho_0)).diagonal().sum())
            new_state = (1-noise)*rho + noise*normalized_rho_0             
            return new_state



"""
>>> this file is a part of tqix: a Toolbox for Quantum in X
                              x: quantum measurement, quantum metrology, 
                                 quantum tomography, and more.
________________________________
>>> copyright (c) 2019 and later
>>> authors: Binho Le
>>> contributors: 
>>> all rights reserved
________________________________
"""

#from numpy import
import numpy as np
from tqix.qobj import *
import math 
from scipy.sparse import csc_matrix

__all__ = ['get_Nds','get_dim','get_num_block','get_array_block',
           'get_jmin','get_jarray','get_marray',
           'get_vidx','get_midx','get_jmm1_idx','get_mm1_idx_max','get_A',
           'get_B','get_D','get_Lambda','get_alpha','dicke_bx','isclose','dicke_bx1','fit'
            ]
        
def get_Nds(d):
    """to get number of qubits from dimension d

    :param d: dimension
    :type d: int
    :return: N
    :rtype: number of qubits
    """    

    d4 = d*4
    root = np.sqrt(d4)
    if int(root + 0.5)**2 == d4: #perfect quare number
        return int(root - 2)
    else:
        return int(np.sqrt(1+d4)-2)
        
def get_dim(N):
    """to get dimension from number of qubits N

    :param N: number of qubits
    :type N: int
    :return: d
    :rtype: dimension
    """     
    
    d = (N/2 + 1)**2 - (N % 2)/4
    return int(d)


def get_num_block(N):
    """to get number of block

    :param N: number of qubits
    :type N: int
    :return: Nbj
    :rtype:  Number of block j
    """    
    return int(N/2 + 1 -1/2*(N % 2))
    
def get_array_block(N):
    """an array of block

    :param N: number of qubits
    :type N: int
    :return: array_block
    :rtype: list
    """    
    # an array of block
    num_block = get_num_block(N)
    array_block = [i * (N+2-i) for i in range(1, num_block+1)]
    return array_block

def get_jmin(N):
    """to get j min

    :param N: number of qubits
    :type N: int
    :return: jmin - min of j
    :rtype: int
    """    
    if N % 2 == 0:
        return 0
    else:
        return 0.5
        
def get_jarray(N):
    """to get array of j from N

    :param N: number of qubits
    :type N: int
    :return: jarray - a array of j
    :rtype: list
    """    
    jarray = np.arange(get_jmin(N),N/2 + 1,1)
    return jarray
       
def get_marray(j):
    """to get array of m from j

    :param j: j index
    :type j: int
    :return: marray - a array of j
    :rtype: list
    """    

    marray = np.arange(-j,j+1,1)
    return marray
          
def get_vidx(j,m):
    """to get index in vector state from j,m

    :param j: j index
    :type j: int
    :param m: m index
    :type m: int
    :return: vidx
    :rtype: int
    """    
    # to get index in vector state from j,m
    return int(j-m)

def get_mm1_idx_max(N):
    """to get mm1 or ik

    :param N: number of qubits
    :type N: int
    :return: [mm1,ik]
    :rtype: list
    """    
   
    # to get mm1 or ik

    ik = {}
    mm1 = {}
    j = N/2

    m_vary = get_marray(j)
    for m in m_vary:
        for m1 in m_vary:
            i, k = j - m, j - m1
            mm1[(i, k)] = (m, m1)
            ik[(m, m1)] = (i, k)
    return [mm1,ik]


def get_midx(N,j,m,m1,block):
    """to get index in density matrix
        ref. qutip

    :param N: number of qubits
    :type N: int
    :param j: j index
    :type j: int
    :param m: m index
    :type m: int
    :param m1: m1 index
    :type m1: int
    :param block: list of blocks
    :type block: list
    :return: (i,k)
    :rtype: tuple
    """     
    # to get index in density matrix
    # ref. qutip
    k = int(j-m1)
    kp = int(j-m)
    block_num = int(N/2 - j) #0,1,2,...
    offset = 0
    
    if block_num > 0:
        offset = block[block_num - 1]
    i = kp + offset
    k = k + offset
    return (i,k)
    
def get_jmm1_idx(N):
    """get index i,k of density matrix from j,m,m1

    and revert j,m,m1 from i,k

    ref. qutip

    :param N: number of qubits
    :type N: int
    :return: [jmm1,ik]
    :rtype: list
    """    

    # get index i,k of density matrix from j,m,m1
    # and revert j,m,m1 from i,k
    # ref. qutip
    
    ik = {}
    jmm1 = {}
    
    block = get_array_block(N)
    j_vary = get_jarray(N)
    for j in j_vary:
        m_vary = get_marray(j)
        for m in m_vary:
            for m1 in m_vary:
                i, k = get_midx(N,j,m,m1,block)
                jmm1[(i, k)] = (j, m, m1)
                ik[(j, m, m1)] = (i, k)
    return [jmm1,ik]

def get_A(j,m,type=""):
    """

    :param j: j index
    :type j: int
    :param m: m index
    :type m: int
    :param type: type of A, defaults to ""
    :type type: str, optional
    :return: A_{type}
    :rtype: float
    """    
    if type == "+":
        return np.sqrt((j-m)*(j+m+1))
    elif type == "-":
        return np.sqrt((j+m)*(j-m+1))
    else:
        return m 

def get_B(j,m,type=""):
    """

    :param j: j index
    :type j: int
    :param m: m index
    :type m: int
    :param type: type of B, defaults to ""
    :type type: str, optional
    :return: B_{type}
    :rtype: float
    """    
    if type == "+":
        return np.sqrt((j-m)*(j-m-1))
    elif type == "-":
        return -np.sqrt((j+m)*(j+m-1))
    else:
        return np.sqrt((j+m)*(j-m))

def get_D(j,m,type=""):
    """

    :param j: j index
    :type j: int
    :param m: m index
    :type m: int 
    :param type: type of D, defaults to ""
    :type type: str, optional
    :return: D_{type}
    :rtype: float
    """    
    if type == "+":
        return -np.sqrt((j+m+1)*(j+m+2))
    if type == "-":
        return np.sqrt((j-m+1)*(j-m+2))
    else:
        return np.sqrt((j+m+1)*(j-m+1))

def get_Lambda(N,j,type=""):
    """

    :param N: number of qubits
    :type N: int
    :param j: j index
    :type j: int
    :param type: type of lambda, defaults to ""
    :type type: str, optional
    :return: Lambda_{type}
    :rtype: float
    """      
    if type == "a":
        return (N/2+1)/(2*j*(j+1))
    elif type == "b":
        return (N/2+j+1)/(2*j*(2*j+1))
    else:
        return (N/2-j)/(2*(j+1)*(2*j+1))

def get_alpha(N,j):
    """

    :param N: number of qubits
    :type N: int
    :param j: j index
    :type j: int
    :return: alpha
    :rtype: float
    """     
    return math.factorial(N)/(math.factorial(N/2-j)*math.factorial(N/2+j))

def dicke_bx(N, jmm1):
    """create a dicke basis follow jmm1
    jmm1 as {(j,m,m1):p}

    :param N: number of qubits
    :type N: int
    :param jmm1: store p value at j,m,m1 indexes
    :type jmm1: dict
    :return: new state
    :rtype: ndarray, tensor, sparse
    """    
    # create a dicke basis follow jmm1
    # jmm1 as {(j,m,m1):p}
    
    dim = get_dim(N)
    rho = np.zeros((dim,dim),dtype = complex)
    ik = get_jmm1_idx(N)[1] # return i,k from jmm1
    for key in jmm1:
        i,k = ik[key]
        rho[i,k] = jmm1[key]
    return csc_matrix(rho)

def dicke_bx1(N,jmm1,ik,dim):
    """create a dicke basis follow jmm1

    jmm1 as {(j,m,m1):p}

    :param N: number of qubits
    :type N: int
    :param jmm1: stores p value at j,m,m1 indexes 
    :type jmm1: dict
    :param ik: stores i,k indexes wrt j,m,m1
    :type ik: dict
    :param dim: dimension of rho 
    :type dim: int
    :return: new state
    :rtype: ndarray, tensor, sparse
    """    
    # create a dicke basis follow jmm1
    # jmm1 as {(j,m,m1):p}

    rho = np.zeros((dim,dim),dtype = complex)
    for key in jmm1:
        i,k = ik[key]
        rho[i,k] = jmm1[key]
    return csc_matrix(rho)

def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """Check if 2 numbers is closed

    :param a: first number 
    :type a: float
    :param b: second number
    :type b: float
    :param rel_tol:  defaults to 1e-09
    :type rel_tol: float, optional
    :param abs_tol: defaults to 0.0
    :type abs_tol: float, optional
    :return: True/False
    :rtype: bool
    """     
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

def fit(objective_function,optimizer,init_params,return_loss_hist=None,loss_break=None,return_time_iters=None):
    """

    :param objective_function: objetive function to minimize
    :type objective_function: func
    :param optimizer: optimizer function 
    :type optimizer: func
    :param init_params: list of initial parameters
    :type init_params: list
    :param return_loss_hist: return history of objective values, defaults to None
    :type return_loss_hist: bool, optional
    :param loss_break: early stopping, defaults to None
    :type loss_break: bool, optional
    :param return_time_iters: return time each iterations, defaults to None
    :type return_time_iters: bool, optional
    :return: output
    :rtype: tuple
    """      
    output = optimizer.optimize(len(init_params), objective_function, initial_point=init_params,
                return_loss_hist=return_loss_hist,loss_break=loss_break,return_time_iters=return_time_iters)
    return output